# -*- coding: utf-8 -*-
#
# Copyright 2015-2022 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
"""
Functions and `dsp` model to model the electric motor in planetary P2 position.
"""
import numpy as np
import schedula as sh
import co2mpas.utils as co2_utl
from co2mpas.defaults import dfl

dsp = sh.BlueDispatcher(
    name='Motor P2 planetary',
    description='Models the planetary motor P2 '
                '(motor acting on the planetary gear-set).'
)


@sh.add_function(dsp, outputs=['motor_p2_planetary_maximum_power'])
def identify_motor_p2_planetary_maximum_power(motor_p2_planetary_powers):
    """
    Identify the maximum power of planetary motor P2 [kW].

    :param motor_p2_planetary_powers:
        Power at planetary motor P2 [kW].
    :type motor_p2_planetary_powers: numpy.array

    :return:
        Maximum power of planetary motor P2 [kW].
    :rtype: float
    """
    from .p4 import identify_motor_p4_maximum_power as func
    return func(motor_p2_planetary_powers)


@sh.add_function(dsp, outputs=['motor_p2_planetary_maximum_torque'])
def identify_motor_p2_planetary_maximum_torque(motor_p2_planetary_torques):
    """
    Identify the maximum torque of planetary motor P2 [N*m].

    :param motor_p2_planetary_torques:
        Torque at planetary motor P2 [N*m].
    :type motor_p2_planetary_torques: numpy.array

    :return:
        Maximum torque of planetary motor P2 [N*m].
    :rtype: float
    """
    from .p4 import identify_motor_p4_maximum_torque as func
    return func(motor_p2_planetary_torques)


@sh.add_function(dsp, outputs=['motor_p2_planetary_maximum_power'])
def calculate_motor_p2_planetary_maximum_power(
        motor_p2_planetary_rated_speed, motor_p2_planetary_maximum_torque):
    """
    Calculate the maximum power of planetary motor P2 [kW].

    :param motor_p2_planetary_rated_speed:
        Rated speed of planetary motor P2 [RPM].
    :type motor_p2_planetary_rated_speed: float

    :param motor_p2_planetary_maximum_torque:
        Maximum torque of planetary motor P2 [N*m].
    :type motor_p2_planetary_maximum_torque: float

    :return:
        Maximum power of planetary motor P2 [kW].
    :rtype: float
    """
    from .p4 import calculate_motor_p4_maximum_power as f
    return f(motor_p2_planetary_rated_speed, motor_p2_planetary_maximum_torque)


@sh.add_function(dsp, outputs=['motor_p2_planetary_rated_speed'])
def calculate_motor_p2_planetary_rated_speed(
        motor_p2_planetary_maximum_power, motor_p2_planetary_maximum_torque):
    """
    Calculate the rated speed of planetary motor P2 [RPM].

    :param motor_p2_planetary_maximum_power:
        Maximum power of planetary motor P2 [kW].
    :type motor_p2_planetary_maximum_power: float

    :param motor_p2_planetary_maximum_torque:
        Maximum torque of planetary motor P2 [N*m].
    :type motor_p2_planetary_maximum_torque: float

    :return:
        Rated speed of planetary motor P2 [RPM].
    :rtype: float
    """
    from .p4 import calculate_motor_p4_rated_speed as func
    return func(
        motor_p2_planetary_maximum_power, motor_p2_planetary_maximum_torque
    )


@sh.add_function(dsp, outputs=['motor_p2_planetary_maximum_torque'])
def calculate_motor_p2_planetary_maximum_torque(
        motor_p2_planetary_maximum_power, motor_p2_planetary_rated_speed):
    """
    Calculate the maximum torque of planetary motor P2 [N*m].

    :param motor_p2_planetary_maximum_power:
        Maximum power of planetary motor P2 [kW].
    :type motor_p2_planetary_maximum_power: float

    :param motor_p2_planetary_rated_speed:
        Rated speed of planetary motor P2 [RPM].
    :type motor_p2_planetary_rated_speed: float

    :return:
        Maximum torque of planetary motor P2 [N*m].
    :rtype: float
    """
    from .p4 import calculate_motor_p4_maximum_torque as f
    return f(motor_p2_planetary_maximum_power, motor_p2_planetary_rated_speed)


@sh.add_function(dsp, outputs=['motor_p2_planetary_maximum_power_function'])
def define_motor_p2_planetary_maximum_power_function(
        motor_p2_planetary_maximum_power, motor_p2_planetary_rated_speed):
    """
    Define the maximum power function of planetary motor P2.

    :param motor_p2_planetary_maximum_power:
        Maximum power of planetary motor P2 [kW].
    :type motor_p2_planetary_maximum_power: float

    :param motor_p2_planetary_rated_speed:
        Rated speed of planetary motor P2 [RPM].
    :type motor_p2_planetary_rated_speed: float

    :return:
        Maximum power function of planetary motor P2.
    :rtype: function
    """
    from .p4 import define_motor_p4_maximum_power_function as f
    func = f(motor_p2_planetary_maximum_power, motor_p2_planetary_rated_speed)

    def _maximum_power_function(speeds):
        return func(np.abs(speeds))

    return _maximum_power_function


@sh.add_function(dsp, outputs=['motor_p2_planetary_maximum_powers'])
def calculate_motor_p2_planetary_maximum_powers(
        motor_p2_planetary_speeds, motor_p2_planetary_maximum_power_function):
    """
    Calculate the maximum power vector of planetary motor P2 [kW].

    :param motor_p2_planetary_speeds:
        Rotating speed of planetary motor P2 [RPM].
    :type motor_p2_planetary_speeds: numpy.array | float

    :param motor_p2_planetary_maximum_power_function:
        Maximum power function of planetary motor P2.
    :type motor_p2_planetary_maximum_power_function: function

    :return:
        Maximum power vector of planetary motor P2 [kW].
    :rtype: numpy.array | float
    """
    from .p4 import calculate_motor_p4_maximum_powers as func
    return func(
        motor_p2_planetary_speeds, motor_p2_planetary_maximum_power_function
    )


@sh.add_function(dsp, outputs=['motor_p2_planetary_speed_ratio'])
def identify_motor_p2_planetary_speed_ratio(
        planetary_speeds_in, motor_p2_planetary_speeds):
    """
    Identifies planetary motor P2 speed ratio.

    :param planetary_speeds_in:
        Planetary speed vector [RPM].
    :type planetary_speeds_in: numpy.array | float

    :param motor_p2_planetary_speeds:
        Rotating speed of planetary motor P2 [RPM].
    :type motor_p2_planetary_speeds: numpy.array | float

    :return:
        Ratio between planetary motor P2 speed and planetary speed [-].
    :rtype: float
    """
    from .p4 import identify_motor_p4_speed_ratio as func
    return func(planetary_speeds_in, motor_p2_planetary_speeds)


dsp.add_data('motor_p2_planetary_speed_ratio', 1, sh.inf(10, 1))


@sh.add_function(dsp, outputs=['motor_p2_planetary_speeds'])
def calculate_motor_p2_planetary_speeds(
        planetary_speeds_in, motor_p2_planetary_speed_ratio):
    """
    Calculates rotating speed of planetary motor P2 [RPM].

    :param planetary_speeds_in:
        Planetary speed vector [RPM].
    :type planetary_speeds_in: numpy.array | float

    :param motor_p2_planetary_speed_ratio:
        Ratio between planetary motor P2 speed and planetary speed [-].
    :type motor_p2_planetary_speed_ratio: float

    :return:
        Rotating speed of planetary motor P2 [RPM].
    :rtype: numpy.array | float
    """
    return planetary_speeds_in * motor_p2_planetary_speed_ratio


@sh.add_function(dsp, inputs_kwargs=True, outputs=['planetary_speeds_in'])
def calculate_planetary_speeds_in(
        motor_p2_planetary_speeds, motor_p2_planetary_speed_ratio=1):
    """
    Calculates Gear box speed vector [RPM].

    :param motor_p2_planetary_speeds:
        Rotating speed of planetary motor P2 [RPM].
    :type motor_p2_planetary_speeds: numpy.array | float

    :param motor_p2_planetary_speed_ratio:
        Ratio between planetary motor P2 speed and planetary speed [-].
    :type motor_p2_planetary_speed_ratio: float

    :return:
        Planetary speed vector [RPM].
    :rtype: numpy.array | float
    """
    return motor_p2_planetary_speeds / motor_p2_planetary_speed_ratio


@sh.add_function(dsp, outputs=['motor_p2_planetary_torques'])
def calculate_motor_p2_planetary_torques(
        motor_p2_planetary_powers, motor_p2_planetary_speeds):
    """
    Calculates torque at planetary motor P2 [N*m].

    :param motor_p2_planetary_powers:
        Power at planetary motor P2 [kW].
    :type motor_p2_planetary_powers: numpy.array | float

    :param motor_p2_planetary_speeds:
        Rotating speed of planetary motor P2 [RPM].
    :type motor_p2_planetary_speeds: numpy.array | float

    :return:
        Torque at planetary motor P2 [N*m].
    :rtype: numpy.array | float
    """
    from ...wheels import calculate_wheel_torques as func
    return func(motor_p2_planetary_powers, motor_p2_planetary_speeds)


@sh.add_function(dsp, outputs=['motor_p2_planetary_powers'])
def calculate_motor_p2_planetary_powers(
        motor_p2_planetary_torques, motor_p2_planetary_speeds):
    """
    Calculates power at planetary motor P2 [kW].

    :param motor_p2_planetary_torques:
        Torque at planetary motor P2 [N*m].
    :type motor_p2_planetary_torques: numpy.array | float

    :param motor_p2_planetary_speeds:
        Rotating speed of planetary motor P2 [RPM].
    :type motor_p2_planetary_speeds: numpy.array | float

    :return:
        Power at planetary motor P2 [kW].
    :rtype: numpy.array | float
    """
    from ...wheels import calculate_wheel_powers as func
    return func(motor_p2_planetary_torques, motor_p2_planetary_speeds)


@sh.add_function(dsp, outputs=['motor_p2_planetary_speeds'])
def calculate_motor_p2_planetary_speeds_v1(
        motor_p2_planetary_powers, motor_p2_planetary_torques):
    """
    Calculates rotating speed of planetary motor P2 [RPM].

    :param motor_p2_planetary_powers:
        Power at planetary motor P2 [kW].
    :type motor_p2_planetary_powers: numpy.array | float

    :param motor_p2_planetary_torques:
        Torque at planetary motor P2 [N*m].
    :type motor_p2_planetary_torques: numpy.array | float

    :return:
        Rotating speed of planetary motor P2 [RPM].
    :rtype: numpy.array | float
    """
    from ...wheels import calculate_wheel_torques as func
    return func(motor_p2_planetary_powers, motor_p2_planetary_torques)


dsp.add_data('motor_p2_planetary_efficiency', 0.9)


@sh.add_function(dsp, outputs=['motor_p2_planetary_electric_powers'])
def calculate_motor_p2_planetary_electric_powers(
        motor_p2_planetary_powers, motor_p2_planetary_efficiency):
    """
    Calculates planetary motor P2 electric power [kW].

    :param motor_p2_planetary_powers:
        Power at planetary motor P2 [kW].
    :type motor_p2_planetary_powers: numpy.array | float

    :param motor_p2_planetary_efficiency:
        Motor P2 efficiency [-].
    :type motor_p2_planetary_efficiency: float

    :return:
        Electric power of planetary motor P2 [kW].
    :rtype: numpy.array | float
    """
    from .p4 import calculate_motor_p4_electric_powers as func
    return func(motor_p2_planetary_powers, motor_p2_planetary_efficiency)


@sh.add_function(dsp, outputs=['motor_p2_planetary_powers'])
def calculate_motor_p2_planetary_powers_v1(
        motor_p2_planetary_electric_powers, motor_p2_planetary_efficiency):
    """
    Calculate planetary motor P2 power from electric power and losses [kW].

    :param motor_p2_planetary_electric_powers:
        Electric power of planetary motor P2 [kW].
    :type motor_p2_planetary_electric_powers: numpy.array | float

    :param motor_p2_planetary_efficiency:
        Motor P2 efficiency [-].
    :type motor_p2_planetary_efficiency: float

    :return:
        Power at planetary motor P2 [kW].
    :rtype: numpy.array | float
    """
    from .p4 import calculate_motor_p4_powers_v1 as f
    return f(motor_p2_planetary_electric_powers, motor_p2_planetary_efficiency)


dsp.add_data('has_motor_p2_planetary', False, sh.inf(10, 3))


@sh.add_function(dsp, outputs=['has_motor_p2_planetary'])
def identify_has_motor_p2(motor_p2_planetary_maximum_power):
    """
    Identify if the vehicle has a planetary motor P2 [kW].

    :param motor_p2_planetary_maximum_power:
        Maximum power of planetary motor P2 [kW].
    :type motor_p2_planetary_maximum_power: float

    :return:
        Has the vehicle a motor in planetary P2?
    :rtype: bool
    """
    from .p4 import identify_has_motor_p4 as func
    return func(motor_p2_planetary_maximum_power)


@sh.add_function(dsp, outputs=['motor_p2_planetary_powers'])
def default_motor_p2_planetary_powers(times, has_motor_p2_planetary):
    """
    Return zero power if the vehicle has not a planetary motor P2 [kW].

    :param times:
        Time vector [s].
    :type times: numpy.array

    :param has_motor_p2_planetary:
        Has the vehicle a motor in planetary P2?
    :type has_motor_p2_planetary: bool

    :return:
        Power at planetary motor P2 [kW].
    :rtype: numpy.array
    """
    from .p4 import default_motor_p4_powers as func
    return func(times, has_motor_p2_planetary)


@sh.add_function(dsp, outputs=['planetary_speeds_in'])
def calculate_planetary_speeds_in(
        engine_speeds_out, final_drive_speeds_in, planetary_ratio):
    """
    Calculates the planetary speed [RPM].

    :param engine_speeds_out:
        Engine speed vector [RPM].
    :type engine_speeds_out: numpy.array

    :param final_drive_speeds_in:
        Final drive speed in [RPM].
    :type final_drive_speeds_in: numpy.array

    :param planetary_ratio:
        Fundamental planetary speed ratio [-].
    :type planetary_ratio: float

    :return:
        Planetary speed vector [RPM].
    :rtype: numpy.array
    """
    r = planetary_ratio
    return engine_speeds_out * (1 + r) - final_drive_speeds_in * r


@sh.add_function(dsp, outputs=['engine_speeds_out'])
def calculate_engine_speeds_out(
        planetary_speeds_in, final_drive_speeds_in, planetary_ratio):
    """
    Calculates the engine speed [RPM].

    :param planetary_speeds_in:
        Planetary speed vector [RPM].
    :type planetary_speeds_in: numpy.array

    :param final_drive_speeds_in:
        Final drive speed in [RPM].
    :type final_drive_speeds_in: numpy.array

    :param planetary_ratio:
        Fundamental planetary speed ratio [-].
    :type planetary_ratio: float

    :return:
        Engine speed vector [RPM].
    :rtype: numpy.array
    """
    if planetary_ratio == -1:
        return sh.NONE
    r = planetary_ratio
    return (planetary_speeds_in + final_drive_speeds_in * r) / (1 + r)


@sh.add_function(dsp, outputs=['final_drive_speeds_in'])
def calculate_final_drive_speeds_in(
        planetary_speeds_in, engine_speeds_out, planetary_ratio):
    """
    Calculates final drive speed [RPM].

    :param planetary_speeds_in:
        Planetary speed vector [RPM].
    :type planetary_speeds_in: numpy.array

    :param engine_speeds_out:
        Engine speed vector [RPM].
    :type engine_speeds_out: numpy.array

    :param planetary_ratio:
        Fundamental planetary speed ratio [-].
    :type planetary_ratio: float

    :return:
        Final drive speed in [RPM].
    :rtype: numpy.array
    """
    if not planetary_ratio:
        return sh.NONE
    r = planetary_ratio
    return (engine_speeds_out * (1 + r) - planetary_speeds_in) / r


@sh.add_function(dsp, outputs=['planetary_ratio'])
def identify_planetary_ratio(
        planetary_speeds_in, engine_speeds_out, final_drive_speeds_in):
    """
    Calculates final drive speed [RPM].

    :param planetary_speeds_in:
        Planetary speed vector [RPM].
    :type planetary_speeds_in: numpy.array

    :param engine_speeds_out:
        Engine speed vector [RPM].
    :type engine_speeds_out: numpy.array

    :param final_drive_speeds_in:
        Final drive speed in [RPM].
    :type final_drive_speeds_in: float

    :return:
        Fundamental planetary speed ratio [-].
    :rtype: numpy.array
    """
    r = planetary_speeds_in - engine_speeds_out
    r /= engine_speeds_out - final_drive_speeds_in
    return co2_utl.reject_outliers(r)


@sh.add_function(dsp, outputs=['planetary_mean_efficiency'])
def default_planetary_mean_efficiency(has_motor_p2_planetary):
    """
    Returns the default planetary mean efficiency [-].

    :param has_motor_p2_planetary:
        Has the vehicle a motor in planetary P2?
    :type has_motor_p2_planetary: bool

    :return:
        Planetary mean efficiency [-].
    :rtype: float
    """
    if has_motor_p2_planetary:
        return dfl.functions.default_planetary_mean_efficiency.efficiency
    return 1


@sh.add_function(dsp, outputs=['planetary_ratio'], weight=sh.inf(10, 5))
def default_planetary_ratio(has_motor_p2_planetary):
    """
    Returns the default fundamental planetary speed ratio [-].

    :param has_motor_p2_planetary:
        Has the vehicle a motor in planetary P2?
    :type has_motor_p2_planetary: bool

    :return:
        Fundamental planetary speed ratio [-].
    :rtype: float
    """
    if has_motor_p2_planetary:
        return dfl.functions.default_planetary_ratio.ratio
    return 0



# -*- coding: utf-8 -*-
#
# Copyright 2015-2022 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
"""
Functions and constants to define the clutch_torque_converter_model selector.
"""
import copy
import schedula as sh
from ...physical.clutch_tc import dsp as _clutch_tc
from ...physical.engine import calculate_engine_speeds_out

#: Model name.
name = 'clutch_torque_converter_model'

#: Parameters that constitute the model.
models = ['clutch_window', 'clutch_speed_model', 'torque_converter_speed_model']

#: Inputs required to run the model.
inputs = [
    'gear_box_speeds_in', 'on_engine', 'idle_engine_speed', 'gear_box_type',
    'gears', 'accelerations', 'times', 'gear_shifts', 'engine_speeds_out_hot',
    'velocities', 'has_torque_converter', 'gear_box_torques_in'
]

#: Relevant outputs of the model.
outputs = ['engine_speeds_out']

#: Targets to compare the outputs of the model.
targets = outputs

#: Extra inputs for the metrics.
metrics_inputs = ['on_engine']


def metric_clutch_torque_converter_model(y_true, y_pred, on_engine):
    """
    Metric for the `clutch_torque_converter_model`.

    :param y_true:
        Reference engine speed vector [RPM].
    :type y_true: numpy.array

    :param y_pred:
        Predicted engine speed vector [RPM].
    :type y_pred: numpy.array

    :param on_engine:
        If the engine is on [-].
    :type on_engine: numpy.array

    :return:
        Error.
    :rtype: float
    """
    from co2mpas.utils import mae
    return float(mae(y_true[on_engine], y_pred[on_engine]))


#: Metrics to compare outputs with targets.
metrics = sh.map_list(targets, metric_clutch_torque_converter_model)

#: Upper score limits to raise the warnings.
up_limit = sh.map_list(targets, 100)

#: Prediction model.
dsp = copy.deepcopy(_clutch_tc).add_function(
    function=calculate_engine_speeds_out,
    inputs=['on_engine', 'idle_engine_speed', 'engine_speeds_out_hot',
            'clutch_tc_speeds_delta'],
    outputs=['engine_speeds_out']
)



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2015-2022 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
"""
It provides functions to perform the hard validation.
"""
import logging
import functools
import numpy as np
import schedula as sh
from co2mpas.defaults import dfl

log = logging.getLogger(__name__)


def check_sign_currents(battery_currents, alternator_currents):
    """
    Checks if battery currents and alternator currents have the right signs.

    :param battery_currents:
        Low voltage battery current vector [A].
    :type battery_currents: numpy.array

    :param alternator_currents:
        Alternator currents [A].
    :type alternator_currents: numpy.array

    :return:
        If battery and alternator currents have the right signs.
    :rtype: (bool, bool)
    """
    #: Maximum allowed positive current for the alternator currents check [A].
    import co2mpas.utils as co2_utl
    max_pos_curr = 1.0
    b_c, a_c = battery_currents, alternator_currents

    a = co2_utl.reject_outliers(a_c, med=np.mean)[0]
    a = a <= max_pos_curr
    c = np.cov(a_c, b_c)[0][1]

    if c < 0:
        x = (a, a)
    elif c == 0:
        if any(b_c):
            x = (co2_utl.reject_outliers(b_c, med=np.mean)[0] <= 0, a)
        else:
            x = (True, a)
    else:
        x = (not a, a)
    return x


# noinspection PyUnusedLocal
def _check_sign_currents(data, *args):
    c = ('service_battery_currents', 'alternator_currents')
    try:
        a = sh.selector(c, data, output_type='list')
        s = check_sign_currents(*a)
        if not all(s):
            s = ' and '.join([k for k, v in zip(c, s) if not v])
            msg = "Probably '{}' have/has the wrong sign!".format(s)
            return c, msg
    except KeyError:  # `c` is not in `data`.
        pass


# noinspection PyUnusedLocal
def _check_acr(data, *args):
    s = ('active_cylinder_ratios', 'engine_has_cylinder_deactivation')
    acr = data.get(s[0], (1,))

    has_acr = data.get(s[1], dfl.values.engine_has_cylinder_deactivation)

    if has_acr and len(acr) <= 1:
        msg = "Please since `engine_has_cylinder_deactivation` is True set " \
              "at least two `active_cylinder_ratios` or set False!"
        return s, msg
    elif not has_acr and len(acr) > 1:
        msg = "Please since there are %d `active_cylinder_ratios` set " \
              "`engine_has_cylinder_deactivation = True` " \
              "or remove the extra ratios!" % len(acr)
        return s, msg


# noinspection PyUnusedLocal
def _check_ki_factor(data, *args):
    s = 'has_periodically_regenerating_systems', 'ki_multiplicative', \
        'ki_additive'

    has_prs = data.get(s[0], dfl.values.has_periodically_regenerating_systems)

    if data.get(s[1], 1) > 1 and data.get(s[2], 0) > 0:
        msg = "Please since `ki_multiplicative` is > 1 and `ki_additive` " \
              "is > 0 set `ki_multiplicative = 1` or set `ki_additive = 0`!"
        return s[1:], msg
    elif not has_prs:
        if data.get(s[1], 1) > 1:
            msg = "Please since `ki_multiplicative` is > 1 set " \
                  "`has_periodically_regenerating_systems = True` or set " \
                  "`ki_multiplicative = 1`!"
            return s, msg
        elif data.get(s[2], 0) > 0:
            msg = "Please since `ki_additive` is > 0 set " \
                  "`has_periodically_regenerating_systems = True` or set " \
                  "`ki_additive = 0`!"
            return s, msg


# noinspection PyUnusedLocal
def _check_prediction_gears_not_mt(data, usage, stage, cycle, *args):
    s = ('gear_box_type', 'gears')
    gear_box_type = data.get(s[0], 'manual')
    if stage == 'prediction' and s[1] in data and gear_box_type != 'manual':
        msg = "`gears` cannot be provided when `gear_box_type` is '%s'." \
              " Hence, remove the `gears` or set `gear_box_type` to manual!"
        return s, msg % gear_box_type


@functools.lru_cache(None)
def _get_engine_model(outputs):
    from ...model.physical.engine import dsp
    return dsp.register(memo={}).shrink_dsp(outputs=outputs)


# noinspection PyUnusedLocal
def _check_lean_burn_tech(data, *args):
    s = ('has_lean_burn', 'ignition_type')
    it = _get_engine_model(s[1:])(data, outputs=s[1:]).get(s[1], None)
    has_lb = data.get(s[0], dfl.values.has_lean_burn)
    if has_lb and it not in ('positive', None):
        msg = "`has_lean_burn` cannot be enable with `ignition_type = '%s'`." \
              "Hence, set `has_lean_burn = False` or " \
              "set `ignition_type = 'positive'`!" % it
        return s, msg


# noinspection PyUnusedLocal
def _check_full_load(data, *args):
    s = ('idle_engine_speed_median', 'full_load_speeds')
    try:
        r = sh.selector(s, _get_engine_model(s)(data, s), output_type='list')
    except KeyError:
        return
    if r[0] < r[1][0]:
        msg = "You have not provided Full Load Curve values below %f RPMs. \n" \
              "This may cause issues in the simulation. Please start from " \
              "idle RPM (%f) or correct the " \
              "`idle_engine_speed_median = full_load_speeds[0]`!"
        return s, msg % (r[1][0], r[0])


# noinspection PyUnusedLocal
def _warn_vva(data, *args):
    s = ('engine_has_variable_valve_actuation', 'ignition_type')
    it = _get_engine_model(s[1:]).dispatch(data, outputs=s[1:]).get(s[1], None)
    has_vva = data.get(s[0], dfl.values.engine_has_variable_valve_actuation)
    if has_vva and it not in ('positive', None):
        msg = "Please, ensure that the input combination " \
              "`engine_has_variable_valve_actuation = True` and " \
              "`ignition_type = '%s'` is correct. If it is intentionally " \
              "added, you can neglect this warning. Otherwise, please " \
              "correct the input setting " \
              "`engine_has_variable_valve_actuation = False` or setting " \
              "`ignition_type = 'positive'`!" % it

        log.warning(msg)


# noinspection PyUnusedLocal
def _check_scr(data, *args):
    s = ('has_selective_catalytic_reduction', 'ignition_type')
    out = _get_engine_model(s[1:]).dispatch(data, outputs=s[1:])
    it = out.get(s[1], None)
    has_scr = data.get(s[0], dfl.values.has_selective_catalytic_reduction)
    if has_scr and it == 'positive':
        msg = "`has_selective_catalytic_reduction` cannot be enable with " \
              "`ignition_type = '%s'`." \
              "Hence, set `has_selective_catalytic_reduction = False` or " \
              "set `ignition_type = 'compression'`!" % it
        return s, msg


# noinspection PyUnusedLocal
def _check_has_torque_converter(data, *args):
    s = ('gear_box_type', 'has_torque_converter')
    if data.get(s[1], False) and data.get(s[0], '') == 'manual':
        msg = "`has_torque_converter` cannot be 'True' when " \
              "`gear_box_type` is 'manual'." \
              "Hence, set `has_torque_converter = False` or " \
              "set `gear_box_type != 'manual'`!"
        return s, msg


# noinspection PyUnusedLocal
def _check_relative_electric_energy_change(data, *args):
    c = ('relative_electric_energy_change', 'transition_cycle_index')
    try:
        reec, n = sh.selector(c, data, output_type='list')
        b = np.array(reec) < .04
        if not (b.shape[0] and b.sum() == 1 and b[-1]):
            msg = '{} must have only the last value <.04!'.format(c[0])
            return c[:-1], msg
        if len(reec) != n + 1:
            msg = '`len({}) != {} + 1`!'.format(*c)
            return c, msg
    except KeyError:  # `c` is not in `data`.
        pass


# noinspection PyUnusedLocal
def _check_gear_box(data, *args):
    c = ('gear_box_type', 'is_hybrid')
    try:
        gear_box_type, is_hybrid = sh.selector(c, data, output_type='list')
        if gear_box_type in ('planetary', 'none') and not is_hybrid:
            msg = f"`gear_box_type` cannot be '{gear_box_type}' when " \
                  "`is_hybrid = False`." \
                  f"Hence, set `gear_box_type != '{gear_box_type}'` or " \
                  "set `is_hybrid = True`!"
            return c, msg
    except KeyError:  # `c` is not in `data`.
        pass


def _hard_validation(data, usage, stage=None, cycle=None, *args):
    if usage in ('input', 'target', 'meta'):
        checks = (
            _check_sign_currents,
            _check_acr,
            _check_ki_factor,
            _check_prediction_gears_not_mt,
            _check_lean_burn_tech,
            _check_full_load,
            _warn_vva,
            _check_scr,
            _check_has_torque_converter,
            _check_relative_electric_energy_change,
            _check_gear_box
        )
        for check in checks:
            c = check(data, usage, stage, cycle, *args)
            if c:
                yield c


def hard_validation(inputs, errors):
    """
    Parse input data for the hard validation.

    :param inputs:
        Input data.
    :type inputs: dict

    :param errors:
        Errors container.
    :type errors: dict

    :return:
        Parsed input data and errors container.
    :rtype: dict, dict
    """
    from schema import SchemaError
    for k, v in sh.stack_nested_keys(inputs, depth=3):
        for c, msg in _hard_validation(v, *k):
            sh.get_nested_dicts(errors, *k)[c] = SchemaError([], [msg])
    return inputs, errors



# -*- coding: utf-8 -*-
#
# Copyright 2015-2022 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
"""
Functions and constants to define the electrics_model selector.
"""
import schedula as sh
import co2mpas.utils as co2_utl
from ._core import define_sub_model
from ...physical.electrics import dsp as _electrics

#: Model name.
name = 'electrics_model'

#: Parameters that constitute the model.
models = [
    'alternator_current_model', 'dcdc_current_model', 'has_energy_recuperation',
    'service_battery_status_model', 'engine_moment_inertia', 'drive_battery_r0',
    'delta_time_engine_starter', 'drive_battery_capacity', 'drive_battery_load',
    'alternator_efficiency', 'service_battery_capacity', 'service_battery_load',
    'drive_battery_n_parallel_cells', 'drive_battery_ocv', 'starter_efficiency',
    'service_battery_initialization_time', 'service_battery_nominal_voltage',
    'drive_battery_nominal_voltage', 'drive_battery_n_series_cells',
    'dcdc_converter_efficiency', 'starter_nominal_voltage',
    'initial_drive_battery_state_of_charge'
]

#: Inputs required to run the model.
inputs = [
    'drive_battery_electric_powers', 'times', 'motive_powers', 'accelerations',
    'on_engine', 'starter_currents', 'initial_service_battery_state_of_charge'
]

#: Relevant outputs of the model.
outputs = [
    'alternator_currents', 'service_battery_currents', 'drive_battery_currents',
    'dcdc_converter_currents', 'service_battery_state_of_charges',
    'drive_battery_state_of_charges'
]
#: Targets to compare the outputs of the model.
targets = outputs

#: Weights coefficients to compute the model score.
weights = sh.map_list(targets, 1, 1, 1, 1, 0, 0)

#: Metrics to compare outputs with targets.
metrics = sh.map_list(targets, *([co2_utl.mae] * 6))

#: Upper score limits to raise the warnings.
up_limit = dict.fromkeys(
    ('alternator_currents', 'service_battery_currents',
     'drive_battery_currents', 'dcdc_converter_currents'), 60
)

#: Prediction model.
# noinspection PyProtectedMember
dsp = sh.Blueprint(_electrics, inputs, outputs, models)._set_cls(
    define_sub_model
)



# -*- coding: utf-8 -*-
#
# Copyright 2015-2022 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
"""
Functions and `dsp` model to processes a CO2MPAS input file.

Sub-Modules:

.. currentmodule:: co2mpas.core

.. autosummary::
    :nosignatures:
    :toctree: core/

    load
    model
    report
    write
"""
import logging
import schedula as sh
import os.path as osp
from co2mpas.utils import check_first_arg
from .load import dsp as _load
from .model import dsp as _model
from .report import dsp as _report
from .write import dsp as _write

log = logging.getLogger(__name__)

dsp = sh.BlueDispatcher(
    name='core',
    description='Processes a CO2MPAS input file.'
)
_cmd_flags = [
    'only_summary', 'hard_validation', 'declaration_mode', 'enable_selector',
    'type_approval_mode', 'encryption_keys', 'sign_key', 'output_template',
    'output_folder', 'encryption_keys_passwords', 'augmented_summary'
]


@sh.add_function(
    dsp, inputs_kwargs=True, inputs_defaults=True, outputs=_cmd_flags
)
def parse_cmd_flags(cmd_flags=None):
    """
    Parses the command line options.

    :param cmd_flags:
        Command line options.
    :type cmd_flags: dict

    :return:
        Default parameters of process model.
    :rtype: tuple
    """
    flags = sh.combine_dicts(cmd_flags or {}, base={
        'only_summary': False,
        'hard_validation': False,
        'declaration_mode': False,
        'enable_selector': False,
        'type_approval_mode': False,
        'encryption_keys': None,
        'sign_key': None,
        'output_template': sh.NONE,
        'encryption_keys_passwords': None,
        'output_folder': './outputs',
        'augmented_summary': False
    })
    flags['declaration_mode'] |= flags['type_approval_mode']
    flags['hard_validation'] |= flags['declaration_mode']
    if flags['declaration_mode'] and not flags['type_approval_mode'] and \
            flags['enable_selector']:
        log.info(
            'Since CO2MPAS is launched in declaration mode the option '
            '--enable-selector is not used.\n'
            'If you want to use it remove -DM from the cmd.'
        )
        flags['enable_selector'] = False
    return sh.selector(_cmd_flags, flags, output_type='list')


dsp.add_dispatcher(
    dsp=_load,
    inputs=(
        'input_file_name', 'hard_validation', 'declaration_mode', 'cmd_flags',
        'type_approval_mode', 'input_file', 'raw_data', 'encryption_keys',
        'sign_key', 'encryption_keys_passwords', 'enable_selector'
    ),
    outputs=('plan', 'flag', 'dice', 'meta', 'base', 'input_file'),
)

# noinspection PyProtectedMember
dsp.add_function(
    function=sh.SubDispatch(_model),
    inputs=['base'],
    outputs=['solution'],
    input_domain=check_first_arg
)


@sh.add_function(dsp, outputs=['output_data'])
def parse_solution(solution):
    """
    Parse the CO2MPAS model solution.

    :param solution:
        CO2MPAS model solution.
    :type solution: schedula.Solution

    :return:
        CO2MPAS outputs.
    :rtype: dict[dict]
    """
    from .model.selector import calibration_cycles
    res = {}
    for k, v in solution.items():
        k = k.split('.')
        sh.get_nested_dicts(res, *k[:-1])[k[-1]] = v
    for k, v in list(sh.stack_nested_keys(res, depth=3)):
        n, k = k[:-1], k[-1]
        if n == ('output', 'calibration') and k in calibration_cycles:
            v = sh.selector(('co2_emission_value',), v, allow_miss=True)
            if v:
                d = sh.get_nested_dicts(res, 'target', 'prediction')
                d[k] = sh.combine_dicts(v, d.get(k, {}))
    from co2mpas.defaults import dfl
    if dfl.functions.parse_solution.CALIBRATION_AS_TARGETS:
        pred = sh.get_nested_dicts(res, 'target', 'prediction')
        pred.update(sh.combine_nested_dicts(sh.get_nested_dicts(
            res, 'output', 'calibration'
        ), pred, depth=2))

    res['pipe'] = solution.pipe

    return res


dsp.add_dispatcher(
    dsp=_report,
    inputs=['output_data', 'augmented_summary'],
    outputs=['report', 'summary'],
)


def check_only_summary(kw):
    """
    Check if `only_summary` is true.

    :param kw:
        `write` dispatcher inputs.
    :type kw: dict

    :return:
        Is `only_summary` true?
    :rtype: bool
    """
    return not kw.get('only_summary', True)


@sh.add_function(dsp, outputs=['vehicle_name'])
def default_vehicle_name(input_file_name):
    """
    Returns the vehicle name.

    :param input_file_name:
        File path.
    :type input_file_name: str

    :return:
        Vehicle name.
    :rtype: str
    """
    return osp.splitext(osp.basename(input_file_name))[0]


dsp.add_dispatcher(
    dsp=_write,
    inputs=[
        'input_file_name', 'input_file', 'base', 'dice', 'meta', 'vehicle_name',
        'output_file_name', 'report', 'flag', 'type_approval_mode', 'timestamp',
        'encryption_keys', 'output_folder', 'sign_key', 'output_template',
        {'only_summary': sh.SINK}
    ],
    outputs=[
        'output_file', 'start_time', 'timestamp', 'output_file_name', sh.SINK
    ],
    input_domain=check_only_summary
)



# -*- coding: utf-8 -*-
#
# Copyright 2015-2022 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
"""
Functions and `dsp` model to model the alternator.
Sub-Modules:

.. currentmodule:: co2mpas.core.model.physical.electrics.motors.alternator

.. autosummary::
    :nosignatures:
    :toctree: alternator/

    current
"""
import numpy as np
import schedula as sh
from co2mpas.defaults import dfl
from .current import dsp as _current

dsp = sh.BlueDispatcher(name='Alternator', description='Models the alternator.')


@sh.add_function(dsp, outputs=['alternator_electric_powers'])
def calculate_alternator_electric_powers(
        alternator_currents, alternator_nominal_voltage):
    """
    Calculates alternator electric powers [kW].

    :param alternator_currents:
        Alternator currents [A].
    :type alternator_currents: numpy.array | float

    :param alternator_nominal_voltage:
        Alternator nominal voltage [V].
    :type alternator_nominal_voltage: float

    :return:
        Alternator electric powers [kW].
    :rtype: numpy.array | float
    """
    return alternator_currents * alternator_nominal_voltage / 1000


@sh.add_function(dsp, outputs=['alternator_currents'])
def calculate_alternator_currents(
        alternator_electric_powers, alternator_nominal_voltage):
    """
    Calculates alternator currents [A].

    :param alternator_electric_powers:
        Alternator power [kW].
    :type alternator_electric_powers: numpy.array | float

    :param alternator_nominal_voltage:
        Alternator nominal voltage [V].
    :type alternator_nominal_voltage: float

    :return:
        Alternator currents [A].
    :rtype: numpy.array | float
    """
    return alternator_electric_powers / alternator_nominal_voltage * 1000


dsp.add_data('alternator_efficiency', dfl.values.alternator_efficiency)


@sh.add_function(dsp, outputs=['alternator_powers'])
def calculate_alternator_powers(
        alternator_electric_powers, alternator_efficiency):
    """
    Calculates alternator power [kW].

    :param alternator_electric_powers:
        Alternator electric power [kW].
    :type alternator_electric_powers: numpy.array | float

    :param alternator_efficiency:
        Alternator efficiency [-].
    :type alternator_efficiency: float

    :return:
        Alternator power [kW].
    :rtype: numpy.array | float
    """
    from ..p4 import calculate_motor_p4_powers_v1 as func
    return func(alternator_electric_powers, alternator_efficiency)


@sh.add_function(dsp, outputs=['alternator_electric_powers'])
def calculate_alternator_electric_powers_v1(
        alternator_powers, alternator_efficiency):
    """
    Calculates alternator electric power [kW].

    :param alternator_powers:
        Alternator power [kW].
    :type alternator_powers: numpy.array | float

    :param alternator_efficiency:
        Alternator efficiency [-].
    :type alternator_efficiency: float

    :return:
        Alternator electric power [kW].
    :rtype: numpy.array | float
    """
    from ..p4 import calculate_motor_p4_electric_powers as func
    return func(alternator_powers, alternator_efficiency)


dsp.add_dispatcher(
    include_defaults=True,
    dsp_id='current_model',
    dsp=_current,
    inputs=(
        'service_battery_state_of_charges', 'service_battery_charging_statuses',
        'service_battery_initialization_time', 'alternator_charging_currents',
        'accelerations', 'on_engine', 'alternator_currents', 'motive_powers',
        'alternator_current_model', 'times',
    ),
    outputs=('alternator_current_model', 'alternator_currents')
)


@sh.add_function(dsp, outputs=['alternator_currents'], weight=sh.inf(10, 3))
def default_alternator_currents(times, is_hybrid):
    """
    Return zero current if the vehicle is hybrid [A].

    :param times:
        Time vector [s].
    :type times: numpy.array

    :param is_hybrid:
        Is the vehicle hybrid?
    :type is_hybrid: bool

    :return:
        Alternator currents [A].
    :rtype: numpy.array
    """
    if is_hybrid:
        return np.zeros_like(times, float)
    return sh.NONE



# -*- coding: utf-8 -*-
#
# Copyright 2015-2022 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
"""
Functions and `dsp` model to model the service battery (low voltage).

Sub-Modules:

.. currentmodule:: co2mpas.core.model.physical.electrics.batteries.service

.. autosummary::
    :nosignatures:
    :toctree: service/

    status
"""
import functools
import numpy as np
import schedula as sh
from co2mpas.defaults import dfl
import co2mpas.utils as co2_utl
from .status import dsp as _status

dsp = sh.BlueDispatcher(
    name='Service Battery',
    description='Models the service battery (e.g., low voltage).'
)


@sh.add_function(dsp, outputs=['service_battery_currents'])
def calculate_service_battery_currents(
        service_battery_electric_powers, service_battery_nominal_voltage):
    """
    Calculate the service battery current vector [A].

    :param service_battery_electric_powers:
        Service battery electric power [kW].
    :type service_battery_electric_powers: numpy.array

    :param service_battery_nominal_voltage:
        Service battery nominal voltage [V].
    :type service_battery_nominal_voltage: float

    :return:
        Service battery current vector [A].
    :rtype: numpy.array
    """
    c = 1000.0 / service_battery_nominal_voltage
    return service_battery_electric_powers * c


@sh.add_function(dsp, outputs=['service_battery_capacity_kwh'])
def calculate_service_battery_capacity_kwh(
        service_battery_capacity, service_battery_nominal_voltage):
    """
    Calculate service battery capacity [kWh].

    :param service_battery_capacity:
        Service battery capacity [Ah].
    :type service_battery_capacity: float

    :param service_battery_nominal_voltage:
        Service battery nominal voltage [V].
    :type service_battery_nominal_voltage: float

    :return:
        Service battery capacity [kWh].
    :rtype: float
    """
    return service_battery_capacity * service_battery_nominal_voltage / 1000.0


@sh.add_function(dsp, outputs=['service_battery_capacity'])
def calculate_service_battery_capacity(
        service_battery_capacity_kwh, service_battery_nominal_voltage):
    """
    Calculate service battery capacity [Ah].

    :param service_battery_capacity_kwh:
        Service battery capacity [kWh].
    :type service_battery_capacity_kwh: float

    :param service_battery_nominal_voltage:
        Service battery nominal voltage [V].
    :type service_battery_nominal_voltage: float

    :return:
        Service battery capacity [Ah].
    :rtype: float
    """
    return service_battery_capacity_kwh * 1000 / service_battery_nominal_voltage


@sh.add_function(dsp, outputs=['service_battery_nominal_voltage'])
def calculate_service_battery_nominal_voltage(
        service_battery_capacity_kwh, service_battery_capacity):
    """
    Calculate service battery nominal voltage [V].

    :param service_battery_capacity_kwh:
        Service battery capacity [kWh].
    :type service_battery_capacity_kwh: float

    :param service_battery_capacity:
        Service battery capacity [Ah].
    :type service_battery_capacity: float

    :return:
        Service battery nominal voltage [V].
    :rtype: float
    """
    return service_battery_capacity_kwh * 1000 / service_battery_capacity


@sh.add_function(dsp, outputs=['service_battery_currents'], weight=sh.inf(1, 0))
def calculate_service_battery_currents_v1(
        service_battery_capacity, times, service_battery_state_of_charges):
    """
    Calculate the service battery current vector [A].

    :param service_battery_capacity:
        Service battery capacity [Ah].
    :type service_battery_capacity: float

    :param times:
        Time vector [s].
    :type times: numpy.array

    :param service_battery_state_of_charges:
        State of charge of the service battery [%].
    :type service_battery_state_of_charges: numpy.array

    :return:
        Service battery current vector [A].
    :rtype: numpy.array
    """
    from scipy.interpolate import UnivariateSpline as Spline
    soc = service_battery_state_of_charges
    ib = Spline(times, soc, w=np.tile(10, times.shape[0])).derivative()(times)
    return ib * (service_battery_capacity * 36.0)


@sh.add_function(dsp, outputs=['service_battery_capacity'])
def identify_service_battery_capacity(
        times, service_battery_currents, service_battery_state_of_charges):
    """
    Identify service battery capacity [Ah].

    :param times:
        Time vector [s].
    :type times: numpy.array

    :param service_battery_currents:
        Service battery current vector [A].
    :type service_battery_currents: numpy.array

    :param service_battery_state_of_charges:
        State of charge of the service battery [%].
    :type service_battery_state_of_charges: numpy.array

    :return:
        Service battery capacity [Ah].
    :rtype: float
    """
    soc = service_battery_state_of_charges
    ib = calculate_service_battery_currents_v1(1, times, soc)
    b = (ib < -dfl.EPS) | (ib > dfl.EPS)
    return co2_utl.reject_outliers(service_battery_currents[b] / ib[b])[0]


@sh.add_function(dsp, outputs=['service_battery_electric_powers'])
def calculate_service_battery_electric_powers(
        service_battery_currents, service_battery_nominal_voltage):
    """
    Calculate the service battery electric power [kW].

    :param service_battery_currents:
        Service battery current vector [A].
    :type service_battery_currents: numpy.array

    :param service_battery_nominal_voltage:
        Service battery nominal voltage [V].
    :type service_battery_nominal_voltage: float

    :return:
        Service battery electric power [kW].
    :rtype: numpy.array
    """
    return service_battery_currents * (service_battery_nominal_voltage / 1000.0)


@sh.add_function(dsp, outputs=['service_battery_electric_powers_supply'])
def calculate_service_battery_electric_powers_supply(
        alternator_electric_powers, dcdc_converter_electric_powers,
        starter_electric_powers):
    """
    Calculate the service battery electric power supply [kW].

    :param alternator_electric_powers:
        Alternator electric power [kW].
    :type alternator_electric_powers: numpy.array

    :param dcdc_converter_electric_powers:
        DC/DC converter electric power [kW].
    :type dcdc_converter_electric_powers: numpy.array

    :param starter_electric_powers:
        Starter electric power [kW].
    :type starter_electric_powers: numpy.array

    :return:
        Service battery electric power supply [kW].
    :rtype: numpy.array
    """
    p = alternator_electric_powers + dcdc_converter_electric_powers
    p += starter_electric_powers
    return p


@sh.add_function(dsp, outputs=['service_battery_electric_powers'], weight=1)
def calculate_service_battery_electric_powers_v1(
        service_battery_loads, service_battery_electric_powers_supply):
    """
    Calculate the service battery electric power [kW].

    :param service_battery_loads:
        Service battery load vector [kW].
    :type service_battery_loads: numpy.array

    :param service_battery_electric_powers_supply:
        Service battery electric power supply [kW].
    :type service_battery_electric_powers_supply: numpy.array

    :return:
        Service battery electric power [kW].
    :rtype: numpy.array
    """
    return service_battery_loads - service_battery_electric_powers_supply


@sh.add_function(dsp, outputs=['initial_service_battery_state_of_charge'])
def identify_initial_service_battery_state_of_charge(
        service_battery_state_of_charges):
    """
    Identify the initial state of charge of service battery [%].

    :param service_battery_state_of_charges:
        State of charge of the service battery [%].
    :type service_battery_state_of_charges: numpy.array

    :return:
        Initial state of charge of the service battery [%].
    :rtype: float
    """
    return service_battery_state_of_charges[0]


@sh.add_function(
    dsp, outputs=['initial_service_battery_state_of_charge'], weight=10
)
def default_initial_service_battery_state_of_charge(cycle_type):
    """
    Return the default initial state of charge of service battery [%].

    :param cycle_type:
        Cycle type (WLTP or NEDC).
    :type cycle_type: str

    :return:
        Initial state of charge of the service battery [%].
    :rtype: float
    """
    d = dfl.functions.default_initial_service_battery_state_of_charge
    return d.initial_state_of_charge[cycle_type]


@sh.add_function(dsp, outputs=['service_battery_state_of_charges'])
def calculate_service_battery_state_of_charges(
        service_battery_capacity, initial_service_battery_state_of_charge,
        times, service_battery_currents):
    """
    Calculates the state of charge of the service battery [%].

    :param service_battery_capacity:
        Service battery capacity [Ah].
    :type service_battery_capacity: float

    :param initial_service_battery_state_of_charge:
        Initial state of charge of the service battery [%].
    :type initial_service_battery_state_of_charge: float

    :param times:
        Time vector [s].
    :type times: numpy.array

    :param service_battery_currents:
        Service battery current vector [A].
    :type service_battery_currents: numpy.array

    :return:
        State of charge of the service battery [%].
    :rtype: numpy.array
    """

    soc = np.empty_like(times, float)
    soc[0] = initial_service_battery_state_of_charge
    bc = (service_battery_currents[:-1] + service_battery_currents[1:])
    bc *= np.diff(times)
    bc /= 2.0 * service_battery_capacity * 36.0

    for i, v in enumerate(bc, 1):
        soc[i] = max(0.0, min(soc[i - 1] + v, 100.0))

    return soc


@sh.add_function(dsp, outputs=['service_battery_loads'])
def calculate_service_battery_loads(
        service_battery_electric_powers,
        service_battery_electric_powers_supply):
    """
    Calculates service battery load vector [kW].

    :param service_battery_electric_powers:
        Service battery electric power [kW].
    :type service_battery_electric_powers: numpy.array

    :param service_battery_electric_powers_supply:
        Service battery electric power supply [kW].
    :type service_battery_electric_powers_supply: numpy.array

    :return:
        Service battery load vector [kW].
    :rtype: numpy.array
    """
    p = service_battery_electric_powers + service_battery_electric_powers_supply
    return p


@sh.add_function(dsp, outputs=['service_battery_loads'])
def calculate_service_battery_loads_v1(on_engine, service_battery_load):
    """
    Calculates service battery load vector [kW].

    :param on_engine:
        If the engine is on [-].
    :type on_engine: numpy.array

    :param service_battery_load:
        Service electric load (engine off and on) [kW].
    :type service_battery_load: float, float

    :return:
        Service battery load vector [kW].
    :rtype: numpy.array
    """
    return np.where(on_engine, *service_battery_load[::-1])


@sh.add_function(dsp, outputs=['service_battery_load'])
def identify_service_battery_load(
        service_battery_loads, engine_powers_out, on_engine):
    """
    Identifies service electric load (engine off and on) [kW].

    :param service_battery_loads:
        Service battery load vector [kW].
    :type service_battery_loads: numpy.array

    :param engine_powers_out:
        Engine power vector [kW].
    :type engine_powers_out: numpy.array

    :param on_engine:
        If the engine is on [-].
    :type on_engine: numpy.array

    :return:
        Service electric load (engine off and on) [kW].
    :rtype: float, float
    """
    rjo, mae = co2_utl.reject_outliers, co2_utl.mae
    p, b = service_battery_loads, engine_powers_out >= -dfl.EPS
    on = min(0.0, co2_utl.reject_outliers(p[on_engine & b], med=np.mean)[0])
    off, b_off = on, b & ~on_engine & (p < 0)
    if b_off.any():
        off = rjo(p[b_off], med=np.mean)[0]
        if on > off:
            p = p[b]
            if mae(p, on) > mae(p, off):
                on = off
            else:
                off = on
    return off, on


@sh.add_function(dsp, outputs=['service_battery_delta_state_of_charge'])
def calculate_service_battery_delta_state_of_charge(
        service_battery_state_of_charges):
    """
    Calculates the overall delta state of charge of the service battery [%].

    :param service_battery_state_of_charges:
        State of charge of the service battery [%].
    :type service_battery_state_of_charges: numpy.array

    :return:
        Overall delta state of charge of the service battery [%].
    :rtype: float
    """
    soc = service_battery_state_of_charges
    return soc[-1] - soc[0]


dsp.add_dispatcher(
    include_defaults=True,
    dsp_id='status_model',
    dsp=_status,
    inputs=(
        'service_battery_start_window_width', 'service_battery_nominal_voltage',
        'service_battery_state_of_charge_balance', 'on_engine', 'accelerations',
        'alternator_electric_powers', 'dcdc_converter_electric_powers', 'times',
        'service_battery_initialization_time', 'service_battery_status_model',
        'service_battery_electric_powers_supply_threshold', 'motive_powers',
        'service_battery_state_of_charges', 'engine_starts', 'is_hybrid',
        'service_battery_state_of_charge_balance_window',
        'service_battery_capacity',
    ),
    outputs=(
        'service_battery_initialization_time', 'service_battery_status_model',
        'service_battery_electric_powers_supply_threshold',
        'service_battery_state_of_charge_balance_window',
        'service_battery_state_of_charge_balance',
        'service_battery_charging_statuses',
    )
)


# noinspection PyMissingOrEmptyDocstring
class ServiceBatteryModel:
    def __init__(self, service_battery_status_model, dcdc_current_model,
                 alternator_current_model, has_energy_recuperation,
                 service_battery_initialization_time, service_battery_load,
                 initial_service_battery_state_of_charge,
                 service_battery_nominal_voltage, service_battery_capacity):
        self.status = functools.partial(
            service_battery_status_model.predict, has_energy_recuperation,
            service_battery_initialization_time
        )
        self.nominal_voltage = service_battery_nominal_voltage
        self.dcdc = dcdc_current_model
        self.alternator = alternator_current_model
        self.current_load = np.divide(
            service_battery_load, service_battery_nominal_voltage / 1e3
        )
        self._d_soc = service_battery_capacity * 36.0 * 2
        self.init_soc = initial_service_battery_state_of_charge
        self.reset()

    # noinspection PyAttributeOutsideInit
    def reset(self):
        self._prev_status = 0
        self._prev_time = 0
        self._prev_current = 0
        self._prev_soc = self.init_soc

    def __call__(self, time, motive_power, acceleration, on_engine,
                 starter_current=0, prev_soc=None, prev_status=None,
                 update=True):
        if prev_status is None:
            prev_status = self._prev_status
        if prev_soc is None:
            prev_soc = self._prev_soc
        alt_c = dcdc_c = .0
        status_ = self.status(time, prev_status, prev_soc, motive_power)
        if status_:
            if status_ == 1:
                dcdc_c = self.dcdc(time, prev_soc, status_)
            if on_engine:
                alt_c = self.alternator(
                    time, prev_soc, status_, motive_power, acceleration
                )
        c = self.current_load[int(on_engine)] - alt_c - dcdc_c - starter_current
        dsoc = (c + self._prev_current) * (time - self._prev_time) / self._d_soc
        soc = max(0.0, min(prev_soc + dsoc, 100.0))

        if update:
            self._prev_status, self._prev_soc = status_, soc
            self._prev_time, self._prev_current = time, c

        return soc, status_, dcdc_c, alt_c


dsp.add_data('has_energy_recuperation', dfl.values.has_energy_recuperation)


@sh.add_function(dsp, outputs=['service_battery_model'])
def define_service_battery_model(
        service_battery_status_model, dcdc_current_model,
        alternator_current_model, has_energy_recuperation,
        service_battery_initialization_time, service_battery_load,
        initial_service_battery_state_of_charge,
        service_battery_nominal_voltage, service_battery_capacity):
    """
    Define a service battery model.

    :param service_battery_status_model:
        A function that predicts the service battery charging status.
    :type service_battery_status_model: BatteryStatusModel

    :param dcdc_current_model:
        DC/DC converter current model.
    :type dcdc_current_model: callable

    :param alternator_current_model:
        Alternator current model.
    :type alternator_current_model: callable

    :param has_energy_recuperation:
        Is the vehicle equipped with any brake energy recuperation technology?
    :type has_energy_recuperation: bool

    :param service_battery_initialization_time:
        Service battery initialization time delta [s].
    :type service_battery_initialization_time: float

    :param service_battery_load:
        Service electric load (engine off and on) [kW].
    :type service_battery_load: float, float

    :param initial_service_battery_state_of_charge:
        Initial state of charge of the service battery [%].
    :type initial_service_battery_state_of_charge: float

    :param service_battery_nominal_voltage:
        Service battery nominal voltage [V].
    :type service_battery_nominal_voltage: float

    :param service_battery_capacity:
        Service battery capacity [Ah].
    :type service_battery_capacity: float

    :return:
        Service battery model.
    :rtype: ServiceBatteryModel
    """
    return ServiceBatteryModel(
        service_battery_status_model, dcdc_current_model,
        alternator_current_model, has_energy_recuperation,
        service_battery_initialization_time, service_battery_load,
        initial_service_battery_state_of_charge,
        service_battery_nominal_voltage, service_battery_capacity
    )



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2015-2022 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
"""
Functions and `dsp` model to define theoretical profiles of WLTP cycle.

Sub-Modules:

.. currentmodule:: co2mpas.core.model.physical.cycle.WLTP

.. autosummary::
    :nosignatures:
    :toctree: WLTP/

    vel
"""
import logging
import copy
import schedula as sh
from ..NEDC import is_manual
from co2mpas.defaults import dfl
from .vel import dsp as _vel

logging.getLogger('wltp.experiment').setLevel(logging.WARNING)
log = logging.getLogger(__name__)

dsp = sh.BlueDispatcher(
    name='WLTP cycle model',
    description='Returns the theoretical times, velocities, and gears of WLTP.'
)

dsp.add_data(
    'initial_temperature', dfl.values.initial_temperature_WLTP,
    description='Initial temperature of the test cell [°C].'
)
dsp.add_data(
    'max_time', dfl.values.max_time_WLTP, 5, description='Maximum time [s].'
)
dsp.add_data('wltp_base_model', copy.deepcopy(dfl.values.wltp_base_model))


@sh.add_function(dsp, outputs=['base_model'])
def define_wltp_base_model(wltp_base_model):
    """
    Defines WLTP base model.

    :param wltp_base_model:
        WLTP base model params.
    :type wltp_base_model: dict

    :return:
        WLTP base model.
    :rtype: dict
    """
    import wltp.model as wltp_mdl
    # noinspection PyProtectedMember
    return sh.combine_dicts(wltp_mdl._get_model_base(), wltp_base_model)


dsp.add_dispatcher(
    include_defaults=True,
    dsp=_vel,
    inputs=(
        'times', 'base_model', 'velocities', 'speed_velocity_ratios',
        'inertial_factor', 'downscale_phases', 'max_velocity',
        'downscale_factor', 'downscale_factor_threshold', 'vehicle_mass',
        'unladen_mass', 'road_loads', 'engine_max_power', 'wltp_class',
        'engine_speed_at_max_power', 'max_speed_velocity_ratio'
    ),
    outputs=(
        'theoretical_motive_powers',
        {'theoretical_velocities': ('theoretical_velocities', 'velocities')}
    )
)


def wltp_gears(
        full_load_curve, velocities, accelerations, motive_powers,
        speed_velocity_ratios, idle_engine_speed, engine_speed_at_max_power,
        engine_max_power, engine_max_speed, base_model, initial_gears=None):
    """
    Returns the gear shifting profile according to WLTP [-].

    :param full_load_curve:
        Vehicle full load curve.
    :type full_load_curve: function

    :param velocities:
        Velocity vector [km/h].
    :type velocities: numpy.array

    :param accelerations:
        Acceleration vector [m/s2].
    :type accelerations: numpy.array

    :param motive_powers:
        Motive power [kW].
    :type motive_powers: numpy.array

    :param speed_velocity_ratios:
        Speed velocity ratios of the gear box [h*RPM/km].
    :type speed_velocity_ratios: dict[int | float]

    :param idle_engine_speed:
        Engine speed idle median and std [RPM].
    :type idle_engine_speed: (float, float)

    :param engine_speed_at_max_power:
        Rated engine speed [RPM].
    :type engine_speed_at_max_power: float

    :param engine_max_power:
        Maximum power [kW].
    :type engine_max_power: float

    :param engine_max_speed:
        Maximum allowed engine speed [RPM].
    :type engine_max_speed: float

    :param base_model:
        WLTP base model params.
    :type base_model: dict

    :param initial_gears:
        Initial gear vector [-].
    :type initial_gears: numpy.array

    :return:
        Gear vector [-].
    :rtype: numpy.array
    """
    import numpy as np
    import wltp.experiment as wltp_exp
    n_min_drive = None
    svr = [v for k, v in sorted(speed_velocity_ratios.items()) if k]
    idle = idle_engine_speed[0]

    n_norm = np.arange(
        0.0, (engine_max_speed - idle) / (engine_speed_at_max_power - idle), .01
    )

    load_curve = {
        'n_norm': n_norm,
        'p_norm': full_load_curve(
            idle + n_norm * (engine_speed_at_max_power - idle)
        ) / engine_max_power
    }

    b = velocities < 0
    if b.any():
        v = velocities.copy()
        v[b] = 0
    else:
        v = velocities

    res = wltp_exp.run_cycle(
        v, accelerations, motive_powers, svr, idle_engine_speed[0], n_min_drive,
        engine_speed_at_max_power, engine_max_power, load_curve, base_model
    )

    if initial_gears:
        gears = initial_gears.copy()
    else:
        # noinspection PyUnresolvedReferences
        gears = res[0]

    # Apply Driveability-rules.
    # noinspection PyUnresolvedReferences
    wltp_exp.applyDriveabilityRules(v, accelerations, gears, res[1], res[-1])

    gears[gears < 0] = 0
    log.warning('The WLTP gear-shift profile generation is for engineering '
                'purposes and the results are by no means valid according to '
                'the legislation.\nActually they are calculated based on a pre '
                'phase-1a version of the GTR spec.\n '
                'Please provide the gear-shifting profile '
                'within `prediction.WLTP` sheet.')
    return gears


dsp.add_function(
    function=sh.add_args(wltp_gears),
    inputs=(
        'gear_box_type', 'full_load_curve', 'velocities', 'accelerations',
        'motive_powers', 'speed_velocity_ratios', 'idle_engine_speed',
        'engine_speed_at_max_power', 'engine_max_power', 'engine_max_speed',
        'base_model'
    ),
    outputs=['gears'],
    input_domain=is_manual,
    weight=sh.inf(2, 10)
)



# -*- coding: utf-8 -*-
#
# Copyright 2015-2022 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
"""
Functions and `dsp` model to model the basic mechanics of the gear box.
"""
import math
import numpy as np
import schedula as sh
from co2mpas.defaults import dfl
import co2mpas.utils as co2_utl

dsp = sh.BlueDispatcher(
    name='mechanical model',
    description='Models the gear box mechanical.'
)


def _correct_gear_shifts(
        times, ratios, gears, velocity_speed_ratios, shift_window=4.0):
    import scipy.optimize as sci_opt
    from . import calculate_gear_shifts
    shifts = calculate_gear_shifts(gears)
    vsr = np.vectorize(lambda v: velocity_speed_ratios.get(v, 0))
    s = len(gears)

    def _err(v, r):
        v = int(v) or 1
        return np.float32(co2_utl.mae(ratios[slice(v - 1, v + 1, 1)], r))

    k = 0
    new_gears = np.zeros_like(gears)
    dt = shift_window / 2
    for i in np.arange(s)[shifts]:
        g = gears[slice(i - 1, i + 1, 1)]
        j = int(i)
        if g[0] != 0 and g[-1] != 0:
            t = times[i]
            n = max(i - (((t - dt) <= times) & (times <= t)).sum(), min(i, k))
            m = min(i + ((t <= times) & (times <= (t + dt))).sum(), s)
            if n + 1 > m:
                # noinspection PyTypeChecker
                j = int(sci_opt.brute(
                    _err, (slice(n, m, 1),), args=(vsr(g),), finish=None)
                )
        x = slice(j - 1, j + 1, 1)
        new_gears[x] = g
        new_gears[k:x.start] = g[0]
        k = x.stop

    new_gears[k:] = new_gears[k - 1]

    return new_gears


dsp.add_data('stop_velocity', dfl.values.stop_velocity)
dsp.add_data('plateau_acceleration', dfl.values.plateau_acceleration)
dsp.add_data('change_gear_window_width', dfl.values.change_gear_window_width)


@sh.add_function(dsp, outputs=['gears'])
def identify_gears(
        times, velocities, accelerations, gear_box_speeds_in,
        velocity_speed_ratios, stop_velocity, plateau_acceleration,
        change_gear_window_width, idle_engine_speed=(0, 0)):
    """
    Identifies gear time series [-].

    :param times:
        Time vector [s].
    :type times: numpy.array

    :param velocities:
        Velocity vector [km/h].
    :type velocities: numpy.array

    :param accelerations:
        Acceleration vector [m/s2].
    :type accelerations: numpy.array

    :param gear_box_speeds_in:
        Gear box speed [RPM].
    :type gear_box_speeds_in: numpy.array

    :param velocity_speed_ratios:
        Constant velocity speed ratios of the gear box [km/(h*RPM)].
    :type velocity_speed_ratios: dict[int | float]

    :param stop_velocity:
        Maximum velocity to consider the vehicle stopped [km/h].
    :type stop_velocity: float

    :param plateau_acceleration:
        Maximum acceleration to be at constant velocity [m/s2].
    :type plateau_acceleration: float

    :param change_gear_window_width:
        Time window used to apply gear change filters [s].
    :type change_gear_window_width: float

    :param idle_engine_speed:
        Engine speed idle median and std [RPM].
    :type idle_engine_speed: (float, float)

    :return:
        Gear vector identified [-].
    :rtype: numpy.array
    """
    with np.errstate(divide='ignore', invalid='ignore'):
        r = velocities / gear_box_speeds_in

    idle = (idle_engine_speed[0] - idle_engine_speed[1],
            idle_engine_speed[0] + idle_engine_speed[1])
    r[gear_box_speeds_in <= idle[0]] = 0

    vsr = velocity_speed_ratios
    g, vsr = np.array([(k, v) for k, v in sorted(vsr.items()) if k != 0]).T
    dr = np.abs(vsr[:, None] - r)
    i, j = np.argmin(dr, 0), np.arange(times.shape[0])
    b = velocities <= vsr[i] * idle[0]
    if idle[1]:
        b |= np.abs(velocities / idle[1] - r) < dr[i, j]
    b = (velocities <= stop_velocity) | (b & (accelerations < 0))
    gear = np.where(b, 0, g[i])
    b = (velocities > stop_velocity) & (accelerations > 0)
    b |= accelerations > plateau_acceleration
    gear[(gear == 0) & b] = 1

    gear = co2_utl.median_filter(times, gear, change_gear_window_width)

    gear = _correct_gear_shifts(times, r, gear, velocity_speed_ratios)

    gear = co2_utl.clear_fluctuations(times, gear, change_gear_window_width)

    return gear.astype(int)


def _shift(a):
    return np.where(np.ediff1d(a.astype(float), [1], [1]) != 0)[0].tolist()


def _calibrate_gsm(
        velocity_speed_ratios, on_engine, anomalies, gear, velocities,
        stop_velocity, idle_engine_speed):
    # noinspection PyProtectedMember
    from .at_gear.cmv import CMV, _filter_gear_shifting_velocity as filter_gs
    idle = idle_engine_speed[0] - idle_engine_speed[1]
    _vsr = sh.combine_dicts(velocity_speed_ratios, base={0: 0})

    limits = {
        0: {False: [0]},
        1: {True: [stop_velocity]},
        max(_vsr): {True: [dfl.INF]}
    }
    shifts = np.unique(sum(map(_shift, (on_engine, anomalies)), []))
    for i, j in co2_utl.pairwise(shifts):
        if on_engine[i:j].all() and not anomalies[i:j].any():
            for v in np.array(list(co2_utl.pairwise(_shift(gear[i:j])))) + i:
                if j != v[1]:
                    v, (g, ng) = velocities[slice(*v)], gear[[v[1] - 1, v[1]]]
                    up = g < ng
                    sh.get_nested_dicts(limits, g, up, default=list).append(
                        v.max() if up else v.min()
                    )

    for k, v in list(limits.items()):
        limits[k] = v.get(False, [_vsr[k] * idle] * 2), v.get(True, [])
    d = {j: i for i, j in enumerate(sorted(limits))}
    gsm = CMV(filter_gs(sh.map_dict(d, limits), stop_velocity))
    gsm.velocity_speed_ratios = sh.selector(gsm, sh.map_dict(d, _vsr))
    gsm.convert(_vsr)
    return gsm


@sh.add_function(dsp, inputs_kwargs=True, outputs=['gears'], weight=10)
@sh.add_function(
    dsp, function_id='identify_gears_v2', outputs=['gears'], weight=20
)
def identify_gears_v1(
        times, velocities, accelerations, on_engine,
        engine_speeds_out, velocity_speed_ratios, stop_velocity,
        plateau_acceleration, change_gear_window_width, idle_engine_speed,
        correct_gear, motive_powers=None):
    """
    Identifies gear time series [-].

    :param times:
        Time vector [s].
    :type times: numpy.array

    :param velocities:
        Velocity vector [km/h].
    :type velocities: numpy.array

    :param accelerations:
        Acceleration vector [m/s2].
    :type accelerations: numpy.array

    :param motive_powers:
        Motive power [kW].
    :type motive_powers: numpy.array

    :param on_engine:
        If the engine is on [-].
    :type on_engine: numpy.array

    :param engine_speeds_out:
        Engine speed [RPM].
    :type engine_speeds_out: numpy.array

    :param velocity_speed_ratios:
        Constant velocity speed ratios of the gear box [km/(h*RPM)].
    :type velocity_speed_ratios: dict[int | float]

    :param stop_velocity:
        Maximum velocity to consider the vehicle stopped [km/h].
    :type stop_velocity: float

    :param plateau_acceleration:
        Maximum acceleration to be at constant velocity [m/s2].
    :type plateau_acceleration: float

    :param change_gear_window_width:
        Time window used to apply gear change filters [s].
    :type change_gear_window_width: float

    :param idle_engine_speed:
        Engine speed idle median and std [RPM].
    :type idle_engine_speed: (float, float)

    :param correct_gear:
        A function to correct the gear predicted.
    :type correct_gear: callable

    :return:
        Gear vector identified [-].
    :rtype: numpy.array
    """
    n = [k for k in velocity_speed_ratios if k != 0]
    if len(n) == 1:
        gears = np.ones_like(times, int) * n[0]
        gears[velocities <= stop_velocity] = 0
        return gears
    with np.errstate(divide='ignore', invalid='ignore'):
        r = velocities / engine_speeds_out

    idle = (idle_engine_speed[0] - idle_engine_speed[1],
            idle_engine_speed[0] + idle_engine_speed[1])
    r[engine_speeds_out <= idle[0]] = 0

    _vsr = sh.combine_dicts(velocity_speed_ratios, base={0: 0})
    g, vsr = np.array([(k, v) for k, v in sorted(_vsr.items())]).T
    dr = np.abs(vsr[:, None] - r)
    i, j = np.argmin(dr, 0), np.arange(times.shape[0])
    b = velocities <= vsr[i] * idle[0]
    if idle[1]:
        b |= np.abs(velocities / idle[1] - r) < dr[i, j]
    b = (velocities <= stop_velocity) | (b & (accelerations < 0))
    gears = np.where(b, 0, g[i]).astype(int)
    gears = co2_utl.median_filter(times, gears, change_gear_window_width)
    gears = _correct_gear_shifts(times, r, gears, velocity_speed_ratios)
    gears = co2_utl.clear_fluctuations(times, gears, change_gear_window_width)
    anomalies = velocities > stop_velocity
    anomalies &= (accelerations > 0) | ~on_engine
    anomalies |= accelerations > plateau_acceleration
    anomalies &= gears == 0
    from ..control.conventional import calculate_engine_speeds_out_hot
    ds = calculate_engine_speeds_out_hot(calculate_gear_box_speeds_in(
        gears, velocities, velocity_speed_ratios, stop_velocity
    ), on_engine, idle_engine_speed) - engine_speeds_out
    i = np.where(on_engine & ~anomalies)[0]
    med = np.nanmedian(ds[i])
    std = 3 * max(50, co2_utl.mad(ds[i], med=med))
    anomalies[i] |= np.abs(ds[i] - med) >= std
    b = (gears == 0) & (velocities <= stop_velocity)
    anomalies[b] = False
    anomalies = co2_utl.clear_fluctuations(
        times, anomalies.astype(int), change_gear_window_width
    )
    for i, j in co2_utl.pairwise(_shift(np.where(b, -1, gears))):
        anomalies[i:j] = anomalies[i:j].mean() > .3

    gsm = _calibrate_gsm(
        velocity_speed_ratios, on_engine, anomalies, gears, velocities,
        stop_velocity, idle_engine_speed
    ).init_gear(
        gears, times, velocities, accelerations, motive_powers,
        correct_gear=correct_gear
    )
    for i, v in enumerate(anomalies):
        if v:
            gears[i] = gsm(i)
    gears = co2_utl.median_filter(times, gears, change_gear_window_width)
    gears = co2_utl.clear_fluctuations(times, gears, change_gear_window_width)
    return gears.astype(int)


@sh.add_function(dsp, outputs=['gear_box_speeds_in'])
def calculate_gear_box_speeds_in_v1(
        gears, gear_box_speeds_out, gear_box_ratios):
    """
    Calculates Gear box speed vector [RPM].

    :param gears:
        Gear vector [-].
    :type gears: numpy.array

    :param gear_box_speeds_out:
        Wheel speed vector [RPM].
    :type gear_box_speeds_out: numpy.array

    :param gear_box_ratios:
        Gear box ratios [-].
    :type gear_box_ratios: dict[int, float | int]

    :return:
        Gear box speed vector [RPM].
    :rtype: numpy.array
    """

    d = {0: 0.0}

    d.update(gear_box_ratios)

    ratios = np.vectorize(lambda k: d[k])(gears)

    return gear_box_speeds_out * ratios


@sh.add_function(dsp, outputs=['gear_box_speeds_in'], weight=25)
def calculate_gear_box_speeds_in(
        gears, velocities, velocity_speed_ratios, stop_velocity):
    """
    Calculates Gear box speed vector [RPM].

    :param gears:
        Gear vector [-].
    :type gears: numpy.array

    :param velocities:
        Velocity vector [km/h].
    :type velocities: numpy.array

    :param velocity_speed_ratios:
        Constant velocity speed ratios of the gear box [km/(h*RPM)].
    :type velocity_speed_ratios: dict[int | float]

    :param stop_velocity:
        Maximum velocity to consider the vehicle stopped [km/h].
    :type stop_velocity: float

    :return:
        Gear box speed vector [RPM].
    :rtype: numpy.array
    """

    speeds = np.array(velocities, dtype=float)
    n = velocities >= stop_velocity
    b = ~n
    for k, r in velocity_speed_ratios.items():
        if r:
            speeds[n & (gears == k)] /= r
        else:
            b |= gears == k

    speeds[b] = 0.0
    return speeds


@sh.add_function(dsp, outputs=['speed_velocity_ratios'])
def calculate_speed_velocity_ratios(
        gear_box_ratios, final_drive_ratios, r_dynamic):
    """
    Calculates speed velocity ratios of the gear box [h*RPM/km].

    :param gear_box_ratios:
        Gear box ratios [-].
    :type gear_box_ratios: dict[int, float | int]

    :param final_drive_ratios:
        Final drive ratios [-].
    :type final_drive_ratios: dict[int, float | int]

    :param r_dynamic:
        Dynamic radius of the wheels [m].
    :type r_dynamic: float

    :return:
        Speed velocity ratios of the gear box [h*RPM/km].
    :rtype: dict
    """

    c = 30 / (3.6 * math.pi * r_dynamic)

    svr = {k: c * v * final_drive_ratios[k] for k, v in gear_box_ratios.items()}

    svr[0] = dfl.INF

    return svr


@sh.add_function(dsp, outputs=['speed_velocity_ratios'], weight=5)
@sh.add_function(
    dsp, inputs=['gears', 'velocities', 'engine_speeds_out', 'stop_velocity'],
    outputs=['speed_velocity_ratios'], weight=10
)
def identify_speed_velocity_ratios(
        gears, velocities, gear_box_speeds_in, stop_velocity):
    """
    Identifies speed velocity ratios from gear vector [h*RPM/km].

    :param gears:
        Gear vector [-].
    :type gears: numpy.array

    :param velocities:
        Velocity vector [km/h].
    :type velocities: numpy.array

    :param gear_box_speeds_in:
        Gear box speed vector [RPM].
    :type gear_box_speeds_in: numpy.array

    :param stop_velocity:
        Maximum velocity to consider the vehicle stopped [km/h].
    :type stop_velocity: float

    :return:
        Speed velocity ratios of the gear box [h*RPM/km].
    :rtype: dict
    """
    with np.errstate(divide='ignore', invalid='ignore'):
        ratios = gear_box_speeds_in / velocities

    ratios[velocities < stop_velocity] = 0

    svr = {k: co2_utl.reject_outliers(ratios[gears == k])[0]
           for k in range(1, int(max(gears)) + 1)
           if k in gears}
    svr[0] = dfl.INF

    return svr


@sh.add_function(dsp, outputs=['velocity_speed_ratios'])
def calculate_velocity_speed_ratios(speed_velocity_ratios):
    """
    Calculates velocity speed (or speed velocity) ratios of the gear box
    [km/(h*RPM) or h*RPM/km].

    :param speed_velocity_ratios:
        Constant speed velocity (or velocity speed) ratios of the gear box
        [h*RPM/km or km/(h*RPM)].
    :type speed_velocity_ratios: dict[int | float]

    :return:
        Constant velocity speed (or speed velocity) ratios of the gear box
        [km/(h*RPM) or h*RPM/km].
    :rtype: dict
    """

    def _inverse(v):
        if v <= 0:
            return dfl.INF
        elif v >= dfl.INF:
            return 0.0
        else:
            return 1 / v

    return {k: _inverse(v) for k, v in speed_velocity_ratios.items()}


@sh.add_function(dsp, outputs=['n_gears'])
@sh.add_function(dsp, inputs=['velocity_speed_ratios'], outputs=['n_gears'])
def identify_n_gears(gear_box_ratios):
    """
    Identify the number of gears [-].

    :param gear_box_ratios:
        Gear box ratios [-].
    :type gear_box_ratios: dict[int, float | int]

    :return:
        Number of gears [-].
    :rtype: int
    """
    return max(gear_box_ratios)


@sh.add_function(dsp, outputs=['velocity_speed_ratios'], weight=48)
def identify_velocity_speed_ratios(
        n_gears, gear_box_speeds_in, velocities, stop_velocity):
    """
    Identifies velocity speed ratios from gear box speed vector [km/(h*RPM)].

    :param n_gears:
        Number of gears [-].
    :type n_gears: int

    :param gear_box_speeds_in:
        Gear box speed [RPM].
    :type gear_box_speeds_in: numpy.array

    :param velocities:
        Velocity vector [km/h].
    :type velocities: numpy.array

    :param stop_velocity:
        Maximum velocity to consider the vehicle stopped [km/h].
    :type stop_velocity: float

    :return:
        Constant velocity speed ratios of the gear box [km/(h*RPM)].
    :rtype: dict
    """
    return identify_velocity_speed_ratios_v1(
        n_gears, gear_box_speeds_in, velocities, (0, 0), stop_velocity
    )


@sh.add_function(dsp, outputs=['velocity_speed_ratios'], weight=49)
def identify_velocity_speed_ratios_v1(
        n_gears, engine_speeds_out, velocities, idle_engine_speed,
        stop_velocity):
    """
    Identifies velocity speed ratios from gear box speed vector [km/(h*RPM)].

    :param n_gears:
        Number of gears [-].
    :type n_gears: int

    :param engine_speeds_out:
        Engine speed [RPM].
    :type engine_speeds_out: numpy.array

    :param velocities:
        Velocity vector [km/h].
    :type velocities: numpy.array

    :param idle_engine_speed:
        Engine speed idle median and std [RPM].
    :type idle_engine_speed: (float, float)

    :param stop_velocity:
        Maximum velocity to consider the vehicle stopped [km/h].
    :type stop_velocity: float

    :return:
        Constant velocity speed ratios of the gear box [km/(h*RPM)].
    :rtype: dict
    """
    import sklearn.cluster as sk_clu

    idle_speed = idle_engine_speed[0] + idle_engine_speed[1]

    b = (engine_speeds_out > idle_speed) & (velocities > stop_velocity)
    x = (velocities[b] / engine_speeds_out[b])[:, None]

    ms = sk_clu.KMeans(n_clusters=int(n_gears), copy_x=False, random_state=0)
    ms.fit(x)

    vsr = {k + 1: v for k, v in enumerate(sorted(ms.cluster_centers_[:, 0]))}

    vsr[0] = 0.0

    return vsr


@sh.add_function(dsp, outputs=['velocity_speed_ratios'], weight=50)
def identify_velocity_speed_ratios_v2(
        gear_box_speeds_in, velocities, stop_velocity):
    """
    Identifies velocity speed ratios from gear box speed vector [km/(h*RPM)].

    :param gear_box_speeds_in:
        Gear box speed [RPM].
    :type gear_box_speeds_in: numpy.array

    :param velocities:
        Velocity vector [km/h].
    :type velocities: numpy.array

    :param stop_velocity:
        Maximum velocity to consider the vehicle stopped [km/h].
    :type stop_velocity: float

    :return:
        Constant velocity speed ratios of the gear box [km/(h*RPM)].
    :rtype: dict
    """
    return identify_velocity_speed_ratios_v3(
        gear_box_speeds_in, velocities, (0, 0), stop_velocity
    )


@sh.add_function(dsp, outputs=['velocity_speed_ratios'], weight=51)
def identify_velocity_speed_ratios_v3(
        engine_speeds_out, velocities, idle_engine_speed, stop_velocity):
    """
    Identifies velocity speed ratios from gear box speed vector [km/(h*RPM)].

    :param engine_speeds_out:
        Engine speed [RPM].
    :type engine_speeds_out: numpy.array

    :param velocities:
        Velocity vector [km/h].
    :type velocities: numpy.array

    :param idle_engine_speed:
        Engine speed idle median and std [RPM].
    :type idle_engine_speed: (float, float)

    :param stop_velocity:
        Maximum velocity to consider the vehicle stopped [km/h].
    :type stop_velocity: float

    :return:
        Constant velocity speed ratios of the gear box [km/(h*RPM)].
    :rtype: dict
    """
    import sklearn.cluster as sk_clu

    idle_speed = idle_engine_speed[0] + idle_engine_speed[1]

    b = (engine_speeds_out > idle_speed) & (velocities > stop_velocity)
    x = (velocities[b] / engine_speeds_out[b])[:, None]

    bandwidth = sk_clu.estimate_bandwidth(x, quantile=0.2)
    ms = sk_clu.MeanShift(bandwidth=bandwidth, bin_seeding=True)
    ms.fit(x)

    vsr = {k + 1: v for k, v in enumerate(sorted(ms.cluster_centers_[:, 0]))}

    vsr[0] = 0.0

    return vsr


@sh.add_function(dsp, outputs=['gear_box_ratios'])
def calculate_gear_box_ratios(
        velocity_speed_ratios, final_drive_ratios, r_dynamic):
    """
    Calculates gear box ratios [-].

    :param velocity_speed_ratios:
        Constant velocity speed ratios of the gear box [km/(h*RPM)].
    :type velocity_speed_ratios: dict[int | float]

    :param final_drive_ratios:
        Final drive ratios [-].
    :type final_drive_ratios: dict[int, float | int]

    :param r_dynamic:
        Dynamic radius of the wheels [m].
    :type r_dynamic: float

    :return:
        Gear box ratios [-].
    :rtype: dict
    """

    c = 30 / (3.6 * math.pi * r_dynamic)

    r = calculate_velocity_speed_ratios(velocity_speed_ratios)

    return {k: v / (c * final_drive_ratios[k]) for k, v in r.items() if k != 0}


@sh.add_function(dsp, outputs=['max_gear'])
def identify_max_gear(speed_velocity_ratios):
    """
    Identifies the maximum gear of the gear box [-].

    :param speed_velocity_ratios:
        Speed velocity ratios of the gear box [h*RPM/km].
    :type speed_velocity_ratios: dict[int | float]

    :return:
        Maximum gear of the gear box [-].
    :rtype: int
    """

    return int(max(speed_velocity_ratios))


@sh.add_function(dsp, outputs=['first_gear_box_ratio'])
def identify_first_gear_box_ratio(gear_box_ratios):
    """
    Identify the gear box ratio of first gear.

    :param gear_box_ratios:
        Gear box ratios [-].
    :type gear_box_ratios: dict[int, float | int]

    :return:
        Gear box ratio of first gear [-].
    :return: float
    """
    return gear_box_ratios.get(1, sh.NONE)


@sh.add_function(dsp, outputs=['last_gear_box_ratio'])
def identify_last_gear_box_ratio(gear_box_ratios):
    """
    Identify the gear box ratio of last gear.

    :param gear_box_ratios:
        Gear box ratios [-].
    :type gear_box_ratios: dict[int, float | int]

    :return:
        Gear box ratio of last gear [-].
    :return: float
    """
    # noinspection PyUnresolvedReferences
    return max(gear_box_ratios.items())[1]


@sh.add_function(dsp, outputs=['engine_speed_at_max_velocity'])
def calculate_engine_speed_at_max_velocity(
        r_dynamic, final_drive_ratios, last_gear_box_ratio, maximum_velocity):
    """
    Calculates the maximum velocity from full load curve.

    :param r_dynamic:
        Dynamic radius of the wheels [m].
    :type r_dynamic: float | numpy.array

    :param final_drive_ratios:
        Final drive ratios [-].
    :type final_drive_ratios: dict[int, float | int]

    :param last_gear_box_ratio:
        Gear box ratio of the last gear [-].
    :type last_gear_box_ratio: float | numpy.array

    :param maximum_velocity:
        Maximum velocity [km/h].
    :type maximum_velocity: float | numpy.array

    :return:
        Engine speed at maximum velocity [RPM].
    :return: float
    """
    speed = last_gear_box_ratio / calculate_last_gear_box_ratio(
        r_dynamic, final_drive_ratios, 1, maximum_velocity
    )
    return speed


@sh.add_function(dsp, outputs=['last_gear_box_ratio'])
def calculate_last_gear_box_ratio(
        r_dynamic, final_drive_ratios, engine_speed_at_max_velocity,
        maximum_velocity):
    """
    Calculates the gear box ratio of the last gear.

    :param r_dynamic:
        Dynamic radius of the wheels [m].
    :type r_dynamic: float

    :param final_drive_ratios:
        Final drive ratios [-].
    :type final_drive_ratios: dict[int, float | int]

    :param engine_speed_at_max_velocity:
        Engine speed at maximum velocity [RPM].
    :type engine_speed_at_max_velocity: float

    :param maximum_velocity:
        Maximum velocity [km/h].
    :type maximum_velocity: float

    :return:
        Gear box ratio of the last gear [-].
    :return: float
    """
    ratio = 3.6 * 2 * np.pi * r_dynamic * engine_speed_at_max_velocity
    # noinspection PyUnresolvedReferences
    ratio /= 60 * max(final_drive_ratios.items())[1] * maximum_velocity
    return ratio


def _calculate_req_power(road_loads, velocity):
    return np.polyval(road_loads[::-1], velocity) * velocity / 3600


@sh.add_function(dsp, outputs=['last_gear_box_ratio'], weight=5)
def calculate_last_gear_box_ratio_v1(
        full_load_curve, final_drive_ratios, road_loads, r_dynamic,
        maximum_velocity):
    """
    Calculates the gear box ratio of the last gear from full load curve.

    :param full_load_curve:
        Vehicle full load curve.
    :type full_load_curve: function

    :param final_drive_ratios:
        Final drive ratios [-].
    :type final_drive_ratios: dict[int, float | int]

    :param road_loads:
        Cycle road loads [N, N/(km/h), N/(km/h)^2].
    :type road_loads: list, tuple

    :param r_dynamic:
        Dynamic radius of the wheels [m].
    :type r_dynamic: float

    :param maximum_velocity:
        Maximum velocity [km/h].
    :type maximum_velocity: float

    :return:
        Gear box ratio of the last gear [-].
    :return: float
    """
    d = dfl.functions.calculate_last_gear_box_ratio_v1
    ratio = np.arange(d.MAX_RATIO, d.MIN_RATIO, -d.DELTA_RATIO)

    p = full_load_curve(calculate_engine_speed_at_max_velocity(
        r_dynamic, final_drive_ratios, ratio, maximum_velocity
    ))

    return ratio[p > _calculate_req_power(road_loads, maximum_velocity)][0]


@sh.add_function(dsp, outputs=['maximum_velocity', 'gear_at_maximum_velocity'])
def calculate_maximum_velocity(
        full_load_curve, road_loads, speed_velocity_ratios):
    """
    Calculates the maximum velocity from full load curve.

    :param full_load_curve:
        Vehicle full load curve.
    :type full_load_curve: function

    :param road_loads:
        Cycle road loads [N, N/(km/h), N/(km/h)^2].
    :type road_loads: list, tuple

    :param speed_velocity_ratios:
        Speed velocity ratios of the gear box [h*RPM/km].
    :type speed_velocity_ratios: dict[int | float]

    :return:
        Maximum velocity and gear at maximum velocity [km/h, -].
    :return: float, int
    """

    d = dfl.functions.calculate_maximum_velocity
    velocity = np.arange(d.MIN_VEL, d.MAX_VEL, d.DELTA_VEL, float)

    g_id, svr = zip(*[(k, v) for k, v in sorted(
        speed_velocity_ratios.items(), reverse=True
    ) if k])

    p = full_load_curve(np.round(np.multiply(velocity[:, None], svr), 1))

    b = (p * d.PREC_FLC) < _calculate_req_power(road_loads, velocity)[:, None]
    # noinspection PyUnresolvedReferences
    velocity = np.repeat(velocity[:, None], b.shape[1], 1)
    velocity[b] = -1

    i, j = np.unravel_index(np.argmax(velocity), velocity.shape)
    return velocity[i, j], g_id[j]


@sh.add_function(dsp, outputs=['maximum_velocity'], weight=10)
def calculate_maximum_velocity_v1(
        r_dynamic, final_drive_ratios, engine_speed_at_max_velocity,
        last_gear_box_ratio):
    """
    Calculates the maximum velocity [km/h].

    :param r_dynamic:
        Dynamic radius of the wheels [m].
    :type r_dynamic: float

    :param final_drive_ratios:
        Final drive ratios [-].
    :type final_drive_ratios: dict[int, float | int]

    :param engine_speed_at_max_velocity:
        Engine speed at maximum velocity [RPM].
    :type engine_speed_at_max_velocity: float

    :param last_gear_box_ratio:
        Gear box ratio of the last gear [-].
    :type last_gear_box_ratio: float

    :return:
        Maximum velocity [km/h].
    :return: float
    """
    vel = calculate_last_gear_box_ratio(
        r_dynamic, final_drive_ratios, engine_speed_at_max_velocity, 1
    ) / last_gear_box_ratio
    return vel


@sh.add_function(dsp, outputs=['maximum_velocity'], weight=15)
def calculate_maximum_velocity_v2(
        full_load_curve, final_drive_ratios, road_loads, r_dynamic,
        last_gear_box_ratio):
    """
    Calculates the maximum velocity from full load curve.

    :param full_load_curve:
        Vehicle full load curve.
    :type full_load_curve: function

    :param final_drive_ratios:
        Final drive ratios [-].
    :type final_drive_ratios: dict[int, float | int]

    :param road_loads:
        Cycle road loads [N, N/(km/h), N/(km/h)^2].
    :type road_loads: list, tuple

    :param r_dynamic:
        Dynamic radius of the wheels [m].
    :type r_dynamic: float

    :param last_gear_box_ratio:
        Gear box ratio of the last gear [-].
    :type last_gear_box_ratio: float

    :return:
        Maximum velocity [km/h].
    :return: float
    """
    d = dfl.functions.calculate_maximum_velocity_v2
    velocity = np.arange(d.MAX_VEL, d.MIN_VEL, -d.DELTA_VEL)

    p = full_load_curve(calculate_engine_speed_at_max_velocity(
        r_dynamic, final_drive_ratios, last_gear_box_ratio, velocity
    ))

    return velocity[p > _calculate_req_power(road_loads, velocity)][0]


@sh.add_function(dsp, outputs=['first_gear_box_ratio'], weight=100)
def calculate_first_gear_box_ratio(
        f0, r_dynamic, engine_max_torque, maximum_vehicle_laden_mass,
        final_drive_ratios):
    """
    Calculates the gear box ratio of the first gear.

    :param f0:
        Rolling resistance force [N] when angle_slope == 0.
    :type f0: float

    :param r_dynamic:
        Dynamic radius of the wheels [m].
    :type r_dynamic: float

    :param engine_max_torque:
        Engine Max Torque [N*m].
    :type engine_max_torque: float

    :param maximum_vehicle_laden_mass:
        Technically permissible maximum laden mass [kg].
    :type maximum_vehicle_laden_mass: float

    :param final_drive_ratios:
        Final drive ratios [-].
    :type final_drive_ratios: dict[int, float | int]

    :return:
        Gear box ratio of the first gear [-].
    :return: float
    """
    d = dfl.functions.calculate_first_gear_box_ratio
    max_torque = engine_max_torque * d.MAX_TORQUE_PERCENTAGE
    slope = np.arctan(d.STARTING_SLOPE)
    ratio = f0 * np.cos(slope)
    ratio += maximum_vehicle_laden_mass * 9.81 * np.sin(slope)
    ratio /= (max_torque * final_drive_ratios[1]) / r_dynamic
    return ratio


@sh.add_function(dsp, outputs=['gear_box_ratios'])
def design_gear_box_ratios(n_gears, first_gear_box_ratio, last_gear_box_ratio):
    """
    Designs the gear box ratios [-].

    :param n_gears:
        Number of gears [-].
    :type n_gears: int

    :param first_gear_box_ratio:
        Gear box ratio of the first gear [-].
    :type first_gear_box_ratio: float

    :param last_gear_box_ratio:
        Gear box ratio of the last gear [-].
    :type last_gear_box_ratio: float

    :return:
        Gear box ratios [-].
    :rtype: dict
    """
    n_gears = int(n_gears)
    d = dfl.functions.design_gear_box_ratios
    f_two, f_tuning = np.asarray(d.f_two), np.asarray(d.f_tuning)

    ix = np.indices((len(f_two), len(f_tuning))).reshape(2, -1).T
    f_two, f_tuning = f_two[ix[:, 0]][:, None], f_tuning[ix[:, 1]][:, None]

    ratios = np.zeros((ix.shape[0], n_gears), float)
    ratios[:, 0], ratios[:, -1] = first_gear_box_ratio, last_gear_box_ratio

    n, fgbr, lgbr = n_gears - 1, first_gear_box_ratio, last_gear_box_ratio
    f_one = f_tuning * (fgbr / (f_two ** (n * (n - 1) / 2))) ** (1 / n)

    n = n_gears - np.arange(2, n_gears)
    ratios[:, 1:-1] = lgbr * (f_one ** n) * (f_two ** (n * (n - 1) / 2))
    dr = np.diff(ratios, axis=1)

    b = np.all(dr < 0, axis=1)
    if b.any():
        ratios, dr = ratios[b], dr[b]

    b = dr[:, 0] < 2
    if b.any():
        ratios, dr = ratios[b], dr[b]

    b = np.all(np.diff(dr, axis=1) > 0, axis=1)
    if b.any():
        ratios, dr = ratios[b], dr[b]

    res = np.linalg.lstsq(
        np.vander(np.arange(n_gears - 1), 3), dr.T, rcond=-1
    )[0]
    b = res[0] < 0

    if b.any():
        ratios, dr, res = ratios[b], dr[b], res[:, b]

    return dict(enumerate(ratios[np.argmin(res[0])], 1))



# -*- coding: utf-8 -*-
#
# Copyright 2015-2022 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
"""
Functions and `dsp` model to model the engine start stop strategy.
"""
import numpy as np
import schedula as sh
from co2mpas.defaults import dfl

dsp = sh.BlueDispatcher(
    name='start_stop', description='Models the engine start/stop strategy.'
)

dsp.add_data('has_start_stop', dfl.values.has_start_stop)


@sh.add_function(dsp, outputs=['start_stop_activation_time'])
def default_start_stop_activation_time(has_start_stop):
    """
    Returns the default start stop activation time threshold [s].

    :return:
        Start-stop activation time threshold [s].
    :rtype: float
    """
    if not has_start_stop or dfl.functions.ENABLE_ALL_FUNCTIONS or \
            dfl.functions.default_start_stop_activation_time.ENABLE:
        return dfl.functions.default_start_stop_activation_time.threshold
    return sh.NONE


# noinspection PyPep8Naming
def _start_stop_model(X):
    off_engine = X[:, 0] <= dfl.functions.StartStopModel.stop_velocity
    off_engine &= X[:, 1] <= dfl.functions.StartStopModel.plateau_acceleration
    return ~off_engine


@sh.add_function(dsp, outputs=['start_stop_model'])
def calibrate_start_stop_model(
        on_engine, velocities, accelerations, times,
        start_stop_activation_time):
    """
    Calibrates an start/stop model to predict if the engine is on.

    :param on_engine:
        If the engine is on [-].
    :type on_engine: numpy.array

    :param velocities:
        Velocity vector [km/h].
    :type velocities: numpy.array

    :param accelerations:
        Acceleration vector [m/s2].
    :type accelerations: numpy.array

    :param times:
        Time vector [s].
    :type times: numpy.array

    :param start_stop_activation_time:
        Start-stop activation time threshold [s].
    :type start_stop_activation_time: float

    :return:
        Start/stop model.
    :rtype: sklearn.tree.DecisionTreeClassifier
    """
    from sklearn.tree import DecisionTreeClassifier
    i = np.searchsorted(times, start_stop_activation_time)
    model = DecisionTreeClassifier(random_state=0, max_depth=3)
    if i >= velocities.shape[0]:
        return _start_stop_model

    model.fit(
        np.column_stack((velocities[i:], accelerations[i:])), on_engine[i:]
    )
    return model.predict


@sh.add_function(dsp, outputs=['correct_start_stop_with_gears'])
def default_correct_start_stop_with_gears(gear_box_type):
    """
    Defines a flag that imposes the engine on when there is a gear > 0.

    :param gear_box_type:
        Gear box type (manual or automatic or cvt).
    :type gear_box_type: str

    :return:
        A flag to impose engine on when there is a gear > 0.
    :rtype: bool
    """

    return gear_box_type == 'manual'


dsp.add_data(
    'min_time_engine_on_after_start', dfl.values.min_time_engine_on_after_start
)


@sh.add_function(dsp, outputs=['on_engine'], weight=sh.inf(1, 0))
def predict_on_engine(
        times, velocities, gears, accelerations, start_stop_model,
        start_stop_activation_time, min_time_engine_on_after_start,
        correct_start_stop_with_gears, has_start_stop):
    """
    Predicts if the engine is on [-].

    :param times:
        Time vector [s].
    :type times: numpy.array

    :param velocities:
        Velocity vector [km/h].
    :type velocities: numpy.array

    :param accelerations:
        Acceleration vector [m/s2].
    :type accelerations: numpy.array

    :param gears:
        Gear vector [-].
    :type gears: numpy.array

    :param start_stop_model:
        Start/stop model.
    :type start_stop_model: callable

    :param start_stop_activation_time:
        Start-stop activation time threshold [s].
    :type start_stop_activation_time: float

    :param min_time_engine_on_after_start:
        Minimum time of engine on after a start [s].
    :type min_time_engine_on_after_start: float

    :param correct_start_stop_with_gears:
        A flag to impose engine on when there is a gear > 0.
    :type correct_start_stop_with_gears: bool

    :param has_start_stop:
        Does the vehicle have start/stop system?
    :type has_start_stop: bool

    :return:
        If the engine is on [-].
    :rtype: numpy.array
    """
    if not has_start_stop:
        return np.ones_like(times, bool)
    on_engine = times <= start_stop_activation_time
    if correct_start_stop_with_gears:
        on_engine |= gears > 0

    b0 = velocities > dfl.functions.StartStopModel.stop_velocity
    b0 |= accelerations > dfl.functions.StartStopModel.plateau_acceleration
    b1 = start_stop_model(np.column_stack((velocities, accelerations)))

    ts = min_time_engine_on_after_start
    # noinspection PyTypeChecker
    for i, (t, on, on_b, on_m) in enumerate(zip(times, on_engine, b0, b1)):
        prev = i == 0 or on_engine[i - 1]
        on = on or ((prev or on_b) and on_m)
        if not on and t <= ts:
            on = True
        if not prev and on:
            ts = t + min_time_engine_on_after_start
        on_engine[i] = on
    return on_engine


@sh.add_function(dsp, outputs=['hybrid_modes'], weight=sh.inf(1, 0))
def default_hybrid_modes(on_engine, gear_box_speeds_in, idle_engine_speed):
    """
    Identify the hybrid mode status (0: EV, 1: Parallel, 2: Serial).

    :param gear_box_speeds_in:
        Gear box speed [RPM].
    :type gear_box_speeds_in: numpy.array | float

    :param on_engine:
        If the engine is on [-].
    :type on_engine: numpy.array, bool

    :param idle_engine_speed:
        Idle engine speed and its standard deviation [RPM].
    :type idle_engine_speed: (float, float)

    :return:
        Hybrid mode status (0: EV, 1: Parallel, 2: Serial).
    :rtype: numpy.array
    """
    modes = on_engine.astype(int)
    modes[on_engine & (gear_box_speeds_in < -np.diff(idle_engine_speed))] = 2
    return modes


@sh.add_function(dsp, inputs_kwargs=True, outputs=['engine_speeds_out_hot'])
def calculate_engine_speeds_out_hot(
        gear_box_speeds_in, on_engine, idle_engine_speed):
    """
    Calculates the engine speed at hot condition [RPM].

    :param gear_box_speeds_in:
        Gear box speed [RPM].
    :type gear_box_speeds_in: numpy.array | float

    :param on_engine:
        If the engine is on [-].
    :type on_engine: numpy.array, bool

    :param idle_engine_speed:
        Idle engine speed and its standard deviation [RPM].
    :type idle_engine_speed: (float, float)

    :return:
        Engine speed at hot condition [RPM].
    :rtype: numpy.array, float
    """
    return np.where(
        on_engine, np.maximum(gear_box_speeds_in, idle_engine_speed[0]), 0
    )



# -*- coding: utf-8 -*-
#
# Copyright 2015-2022 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
"""
Functions and `dsp` model to model the after treatment.
"""
import numpy as np
import schedula as sh
import co2mpas.utils as co2_utl
from co2mpas.defaults import dfl

dsp = sh.BlueDispatcher(
    name='After treatment', description='Models the after treatment.'
)


@sh.add_function(dsp, inputs=[
    'times', 'velocities', 'engine_speeds_base', 'gear_box_speeds_in', 'gears',
    'stop_velocity', 'min_engine_on_speed', 'on_engine', 'idle_engine_speed'
], outputs=['on_idle'])
@sh.add_function(dsp, outputs=['on_idle'])
def identify_on_idle(
        times, velocities, engine_speeds_out_hot, gear_box_speeds_in, gears,
        stop_velocity, min_engine_on_speed, on_engine, idle_engine_speed):
    """
    Identifies when the engine is on idle [-].

    :param times:
        Time vector [s].
    :type times: numpy.array

    :param velocities:
        Velocity vector [km/h].
    :type velocities: numpy.array

    :param engine_speeds_out_hot:
        Engine speed at hot condition [RPM].
    :type engine_speeds_out_hot: numpy.array

    :param gear_box_speeds_in:
        Gear box speed [RPM].
    :type gear_box_speeds_in: numpy.array

    :param gears:
        Gear vector [-].
    :type gears: numpy.array

    :param stop_velocity:
        Maximum velocity to consider the vehicle stopped [km/h].
    :type stop_velocity: float

    :param min_engine_on_speed:
        Minimum engine speed to consider the engine to be on [RPM].
    :type min_engine_on_speed: float

    :param on_engine:
        If the engine is on [-].
    :type on_engine: numpy.array

    :param idle_engine_speed:
        Engine speed idle median and std [RPM].
    :type idle_engine_speed: (float, float)

    :return:
        If the engine is on idle [-].
    :rtype: numpy.array
    """
    # noinspection PyProtectedMember
    from .gear_box.mechanical import _shift
    b = engine_speeds_out_hot > min_engine_on_speed
    b &= (gears == 0) | (velocities <= stop_velocity)

    on_idle = np.zeros_like(times, int)
    i = np.where(on_engine)[0]
    ds = np.abs(gear_box_speeds_in[i] - engine_speeds_out_hot[i])
    on_idle[i[ds > idle_engine_speed[1]]] = 1
    on_idle = co2_utl.median_filter(times, on_idle, 4)
    on_idle[b] = 1
    for i, j in co2_utl.pairwise(_shift(on_idle)):
        if not on_idle[i] and times[j - 1] - times[i] <= 2:
            on_idle[i:j] = 1
    return co2_utl.clear_fluctuations(times, on_idle, 4).astype(bool)


@sh.add_function(
    dsp, inputs_kwargs=True, outputs=['after_treatment_warm_up_phases']
)
def identify_after_treatment_warm_up_phases(
        times, engine_speeds_out, engine_speeds_out_hot, on_idle, on_engine,
        idle_engine_speed, velocities, engine_starts, stop_velocity,
        is_hybrid=False):
    """
    Identifies when engine speed is affected by the after treatment warm up [-].

    :param times:
        Time vector [s].
    :type times: numpy.array

    :param engine_speeds_out:
        Engine speed [RPM].
    :type engine_speeds_out: numpy.array

    :param engine_speeds_out_hot:
        Engine speed at hot condition [RPM].
    :type engine_speeds_out_hot: numpy.array

    :param on_idle:
        If the engine is on idle [-].
    :type on_idle: numpy.array

    :param on_engine:
        If the engine is on [-].
    :type on_engine: numpy.array

    :param idle_engine_speed:
        Engine speed idle median and std [RPM].
    :type idle_engine_speed: (float, float)

    :param velocities:
        Velocity vector [km/h].
    :type velocities: numpy.array

    :param engine_starts:
        When the engine starts [-].
    :type engine_starts: numpy.array

    :param stop_velocity:
        Maximum velocity to consider the vehicle stopped [km/h].
    :type stop_velocity: float

    :param is_hybrid:
        Is the vehicle hybrid?
    :type is_hybrid: bool

    :return:
        Phases when engine speed is affected by the after treatment warm up [-].
    :rtype: numpy.array
    """
    from .control import identify_engine_starts
    i, phases = np.where(on_idle)[0], np.zeros_like(times, int)
    start = engine_starts.copy()
    if is_hybrid:
        with np.errstate(divide='ignore', invalid='ignore'):
            r = engine_speeds_out[i] / engine_speeds_out_hot[i]
            b = ~co2_utl.get_inliers(r, 2, np.nanmedian, co2_utl.mad)[0]
        phases[i[b]] = 1
    else:
        ds = np.abs(engine_speeds_out[i] - engine_speeds_out_hot[i])
        phases[i[ds > idle_engine_speed[1]]] = 1
        start |= identify_engine_starts(velocities > stop_velocity)
    for i, j in np.searchsorted(times, times[start, None] + [-2, 5 + dfl.EPS]):
        phases[i:j], start[i:j] = 0, True
    phases = co2_utl.median_filter(times, phases, 4)
    phases = co2_utl.clear_fluctuations(times, phases, 4).astype(bool)
    indices = co2_utl.index_phases(phases)
    if is_hybrid:
        indices = indices[(np.diff(times[indices], axis=1) > 10).ravel()][:1]
    else:
        b, f = identify_engine_starts(~on_engine), False
        for i, j in np.searchsorted(times, times[b, None] + [-5, 2 + dfl.EPS]):
            b[i:j] = f
            f = True
        b = (on_idle & ~start) | b | (times > np.min(times[on_engine]) + 200)
        indices = indices[:co2_utl.argmax(b[indices[:, 1]])]
    phases[:], n = False, len(times)
    for i, j in indices:
        while i and on_idle.take(i - 1, mode='clip'):
            i -= 1
        phases[i:j + 1] = True
    return phases


@sh.add_function(
    dsp, inputs_kwargs=True, outputs=['after_treatment_warm_up_duration']
)
def identify_after_treatment_warm_up_duration(
        times, after_treatment_warm_up_phases, is_hybrid=False):
    """
    Identify after treatment warm up duration [s].

    :param times:
        Time vector [s].
    :type times: numpy.array

    :param after_treatment_warm_up_phases:
        Phases when engine speed is affected by the after treatment warm up [-].
    :type after_treatment_warm_up_phases: numpy.array

    :param is_hybrid:
        Is the vehicle hybrid?
    :type is_hybrid: bool

    :return:
        After treatment warm up duration [s].
    :rtype: float
    """
    i = co2_utl.index_phases(after_treatment_warm_up_phases)
    if i.shape[0]:
        if not is_hybrid:
            return float(np.diff(times[i.ravel()[[0, -1]]]))
        return float(np.mean(np.diff(times[i], axis=1)))
    return .0


@sh.add_function(dsp, outputs=['after_treatment_warm_up_duration'])
def default_after_treatment_warm_up_duration(is_hybrid):
    """
    Returns the default after treatment warm up duration [s].

    :param is_hybrid:
        Is the vehicle hybrid?
    :type is_hybrid: bool

    :return:
        After treatment warm up duration [s].
    :rtype: float
    """
    if is_hybrid:
        return sh.NONE
    return dfl.functions.default_after_treatment_warm_up_duration.duration


@sh.add_function(dsp, outputs=['after_treatment_cooling_duration'])
def default_after_treatment_cooling_duration(is_hybrid):
    """
    Returns the default after treatment cooling duration [s].

    :param is_hybrid:
        Is the vehicle hybrid?
    :type is_hybrid: bool

    :return:
        After treatment cooling duration [s].
    :rtype: float
    """
    if is_hybrid:
        return sh.NONE
    return dfl.functions.default_after_treatment_cooling_duration.duration


@sh.add_function(
    dsp, inputs_kwargs=True, outputs=['after_treatment_cooling_duration']
)
def identify_after_treatment_cooling_duration(
        times, after_treatment_warm_up_phases, is_hybrid=False):
    """
    Identify after treatment cooling duration [s].

    :param times:
        Time vector [s].
    :type times: numpy.array

    :param after_treatment_warm_up_phases:
        Phases when engine speed is affected by the after treatment warm up [-].
    :type after_treatment_warm_up_phases: numpy.array

    :param is_hybrid:
        Is the vehicle hybrid?
    :type is_hybrid: bool

    :return:
        After treatment cooling duration [s].
    :rtype: float
    """
    i = co2_utl.index_phases(after_treatment_warm_up_phases)
    if is_hybrid and i.shape[0] > 2:
        return float(np.mean(np.diff(times[i].ravel())[1::2]))
    return float('inf')


@sh.add_function(dsp, outputs=['after_treatment_speeds_delta'])
def identify_after_treatment_speeds_delta(
        after_treatment_warm_up_phases, engine_speeds_out,
        engine_speeds_out_hot):
    """
    Identifies the Engine speed delta due to the after treatment warm up [RPM].

    :param after_treatment_warm_up_phases:
        Phases when engine speed is affected by the after treatment warm up [-].
    :type after_treatment_warm_up_phases: numpy.array

    :param engine_speeds_out:
        Engine speed [RPM].
    :type engine_speeds_out: numpy.array

    :param engine_speeds_out_hot:
        Engine speed at hot condition [RPM].
    :type engine_speeds_out_hot: numpy.array

    :return:
        Engine speed delta due to the after treatment warm up [RPM].
    :rtype: numpy.array
    """
    speeds = np.zeros_like(engine_speeds_out, dtype=float)
    b = after_treatment_warm_up_phases
    speeds[b] = np.maximum(0, engine_speeds_out[b] - engine_speeds_out_hot[b])
    return speeds


@sh.add_function(dsp, outputs=['engine_speeds_base'])
def calculate_engine_speeds_base(
        engine_speeds_out_hot, after_treatment_speeds_delta):
    """
    Calculate base engine speed (i.e., without clutch/TC effect) [RPM].

    :param engine_speeds_out_hot:
        Engine speed at hot condition [RPM].
    :type engine_speeds_out_hot: numpy.array

    :param after_treatment_speeds_delta:
        Engine speed delta due to the after treatment warm up [RPM].
    :type after_treatment_speeds_delta: numpy.array

    :return:
        Base engine speed (i.e., without clutch/TC effect) [RPM].
    :rtype: numpy.array
    """
    return engine_speeds_out_hot + after_treatment_speeds_delta


@sh.add_function(
    dsp, inputs_kwargs=True, outputs=['after_treatment_speed_model']
)
def calibrate_after_treatment_speed_model(
        times, after_treatment_warm_up_phases, after_treatment_speeds_delta,
        is_hybrid=False):
    """
    Calibrates the engine after treatment speed model.

    :param times:
        Time vector [s].
    :type times: numpy.array

    :param after_treatment_warm_up_phases:
        Phases when engine speed is affected by the after treatment warm up [-].
    :type after_treatment_warm_up_phases: numpy.array

    :param after_treatment_speeds_delta:
        Engine speed delta due to the after treatment warm up [RPM].
    :type after_treatment_speeds_delta: numpy.array

    :param is_hybrid:
        Is the vehicle hybrid?
    :type is_hybrid: bool

    :return:
        After treatment speed model.
    :rtype: function
    """
    if after_treatment_warm_up_phases.any():
        from sklearn.isotonic import IsotonicRegression
        x, y, model = [], [], IsotonicRegression(increasing=False)
        for i, j in co2_utl.index_phases(after_treatment_warm_up_phases):
            x.extend(times[i:j + 1] - (times[i] if is_hybrid else 0.0))
            y.extend(after_treatment_speeds_delta[i:j + 1])
        # noinspection PyUnresolvedReferences
        return model.fit(x, y).predict


@sh.add_function(
    dsp, inputs_kwargs=True, outputs=['after_treatment_power_model']
)
def calibrate_after_treatment_power_model(
        times, after_treatment_warm_up_phases, engine_powers_out,
        is_hybrid=False):
    """
    Calibrates the engine after treatment speed model.

    :param times:
        Time vector [s].
    :type times: numpy.array

    :param after_treatment_warm_up_phases:
        Phases when engine speed is affected by the after treatment warm up [-].
    :type after_treatment_warm_up_phases: numpy.array

    :param engine_powers_out:
        Engine power vector [kW].
    :type engine_powers_out: numpy.array

    :param is_hybrid:
        Is the vehicle hybrid?
    :type is_hybrid: bool

    :return:
        After treatment speed model.
    :rtype: function
    """
    if after_treatment_warm_up_phases.any():
        from sklearn.isotonic import IsotonicRegression
        x, y = [], []
        for i, j in co2_utl.index_phases(after_treatment_warm_up_phases):
            t = times[i:j + 1] - (times[i] if is_hybrid else 0.0)
            x.extend(t)
            y.extend(co2_utl.median_filter(t, engine_powers_out[i:j + 1], 4))
        # noinspection PyUnresolvedReferences
        return IsotonicRegression().fit(x, np.maximum(0, y)).predict


@sh.add_function(dsp, outputs=['after_treatment_warm_up_phases'])
def predict_after_treatment_warm_up_phases(
        after_treatment_warm_up_duration, after_treatment_cooling_duration,
        times, on_engine, is_cycle_hot):
    """
    Calculates the engine speed delta due to the after treatment warm up [RPM].

    :param after_treatment_warm_up_duration:
        After treatment warm up duration [s].
    :type after_treatment_warm_up_duration: float

    :param after_treatment_cooling_duration:
        After treatment cooling duration [s].
    :type after_treatment_cooling_duration: float

    :param times:
        Time vector [s].
    :type times: numpy.array

    :param on_engine:
        If the engine is on [-].
    :type on_engine: numpy.array

    :param is_cycle_hot:
        Is an hot cycle?
    :type is_cycle_hot: bool

    :return:
        Phases when engine speed is affected by the after treatment warm up [-].
    :rtype: numpy.array
    """
    phases = np.zeros_like(times, bool)
    if after_treatment_warm_up_duration and on_engine.any():
        indices = co2_utl.index_phases(on_engine)
        indices = indices[np.append(not is_cycle_hot, np.diff(
            times[indices].ravel()
        )[1::2] > after_treatment_cooling_duration)]
        for i, j in indices:
            t = times[i:j + 1] - times[i]
            phases[i:j + 1] = t < after_treatment_warm_up_duration
    return phases


@sh.add_function(dsp, outputs=['after_treatment_speeds_delta'])
def predict_after_treatment_speeds_delta(
        after_treatment_speed_model, times, after_treatment_warm_up_phases,
        on_idle, is_hybrid):
    """
    Predicts the engine speed delta due to the after treatment warm up [RPM].

    :param after_treatment_speed_model:
        After treatment speed model.
    :type after_treatment_speed_model: function

    :param times:
        Time vector [s].
    :type times: numpy.array

    :param after_treatment_warm_up_phases:
        Phases when engine speed is affected by the after treatment warm up [-].
    :type after_treatment_warm_up_phases: numpy.array

    :param on_idle:
        If the engine is on idle [-].
    :type on_idle: numpy.array

    :param is_hybrid:
        Is the vehicle hybrid?
    :type is_hybrid: bool

    :return:
        Engine speed delta due to the after treatment warm up [RPM].
    :rtype: numpy.array
    """
    ds = np.zeros_like(times, float)
    if after_treatment_speed_model:
        for i, j in co2_utl.index_phases(after_treatment_warm_up_phases):
            ds[i:j + 1] = after_treatment_speed_model(times[i:j + 1] - times[i])
        if not is_hybrid:
            ds[~on_idle] = 0
    return np.nan_to_num(ds)



# -*- coding: utf-8 -*-
#
# Copyright 2015-2022 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
"""
Functions and `dsp` model to model the CVT.
"""
import numpy as np
import schedula as sh
from co2mpas.defaults import dfl

dsp = sh.BlueDispatcher(name='CVT model', description='Models the gear box.')


@sh.add_function(dsp, outputs=['correct_gear'])
def default_correct_gear():
    """
    Returns a fake function to correct the gear.

    :return:
        A function to correct the predicted gear.
    :rtype: callable
    """
    from .at_gear import CorrectGear
    return CorrectGear()


# noinspection PyMissingOrEmptyDocstring,PyPep8Naming,PyUnusedLocal
class CVT:
    def __init__(self):
        # noinspection PyProtectedMember
        from ..engine._thermal import _XGBRegressor
        self.base_model = _XGBRegressor
        self.model = None

    def fit(self, on_engine, gear_box_speeds_in, velocities, accelerations,
            gear_box_powers_out):
        b = on_engine
        X = np.column_stack((velocities, accelerations, gear_box_powers_out))[b]
        y = gear_box_speeds_in[b]

        # noinspection PyArgumentEqualDefault
        self.model = self.base_model(
            random_state=0,
            max_depth=3,
            n_estimators=int(min(300.0, 0.25 * (len(y) - 1))),
            objective='reg:squarederror'
        )
        self.model.fit(X, y)
        self.model.cache_params()
        return self

    # noinspection PyUnusedLocal
    @staticmethod
    def init_gear(*args, **kwargs):
        return lambda *a, **kw: 1

    def predict(self, velocities, accelerations, gear_box_powers_out):
        X = np.column_stack((velocities, accelerations, gear_box_powers_out))
        return self.model.predict(X)

    def init_speed(self, stop_velocity, gears, velocities, accelerations,
                   gear_box_powers_out):
        predict, x = self.model.predict, np.empty((1, 3), float)
        powers = gear_box_powers_out

        def _next(i):
            x[:, :] = velocities[i], accelerations[i], powers[i]
            return predict(x)

        return _next


@sh.add_function(dsp, outputs=['CVT'])
@sh.add_function(
    dsp,
    inputs=[
        'on_engine', 'engine_speeds_out', 'velocities', 'accelerations',
        'gear_box_powers_out'
    ],
    outputs=['CVT'], weight=50
)
def calibrate_cvt(
        on_engine, gear_box_speeds_in, velocities, accelerations,
        gear_box_powers_out):
    """
    Calibrates a model for continuously variable transmission (CVT).

    :param on_engine:
        If the engine is on [-].
    :type on_engine: numpy.array

    :param gear_box_speeds_in:
        Gear box speed [RPM].
    :type gear_box_speeds_in: numpy.array

    :param velocities:
        Vehicle velocity [km/h].
    :type velocities: numpy.array

    :param accelerations:
        Vehicle acceleration [m/s2].
    :type accelerations: numpy.array

    :param gear_box_powers_out:
        Gear box power vector [kW].
    :type gear_box_powers_out: numpy.array

    :return:
        Continuously variable transmission model.
    :rtype: callable
    """
    model = CVT().fit(
        on_engine, gear_box_speeds_in, velocities, accelerations,
        gear_box_powers_out
    )

    return model


dsp.add_data('max_gear', 1)


@sh.add_function(dsp, outputs=['gears'])
def predict_gears(velocities):
    """
    Predicts gear vector [-].

    :param velocities:
        Vehicle velocity [km/h].
    :type velocities: numpy.array

    :return:
        Gear vector [-].
    :rtype: numpy.array
    """

    return np.ones_like(velocities, dtype=int)


@sh.add_function(
    dsp,
    inputs=['CVT', 'velocities', 'accelerations', 'gear_box_powers_out'],
    outputs=['gear_box_speeds_in'],
    out_weight={'gear_box_speeds_in': 10}
)
def predict_gear_box_speeds_in(
        cvt, velocities, accelerations, gear_box_powers_out):
    """
    Predicts gear box speed vector, gear vector, and maximum gear [RPM, -, -].

    :param cvt:
        Continuously variable transmission model.
    :type cvt: callable

    :param velocities:
        Vehicle velocity [km/h].
    :type velocities: numpy.array

    :param accelerations:
        Vehicle acceleration [m/s2].
    :type accelerations: numpy.array

    :param gear_box_powers_out:
        Gear box power vector [kW].
    :type gear_box_powers_out: numpy.array

    :return:
        Gear box speed vector, gear vector, and maximum gear [RPM, -, -].
    :rtype: numpy.array, numpy.array, int
    """

    return cvt.predict(velocities, accelerations, gear_box_powers_out)


dsp.add_data('stop_velocity', dfl.values.stop_velocity)


@sh.add_function(dsp, outputs=['max_speed_velocity_ratio'])
def identify_max_speed_velocity_ratio(
        velocities, engine_speeds_out, idle_engine_speed, stop_velocity):
    """
    Identifies the maximum speed velocity ratio of the gear box [h*RPM/km].

    :param velocities:
        Vehicle velocity [km/h].
    :type velocities: numpy.array

    :param engine_speeds_out:
        Engine speed [RPM].
    :type engine_speeds_out: numpy.array

    :param idle_engine_speed:
        Engine speed idle median and std [RPM].
    :type idle_engine_speed: (float, float)

    :param stop_velocity:
        Maximum velocity to consider the vehicle stopped [km/h].
    :type stop_velocity: float

    :return:
        Maximum speed velocity ratio of the gear box [h*RPM/km].
    :rtype: float
    """

    b = (velocities > stop_velocity)
    b &= (engine_speeds_out > idle_engine_speed[0])
    return max(engine_speeds_out[b] / velocities[b])



# -*- coding: utf-8 -*-
#
# Copyright 2015-2022 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
"""
Functions to write outputs on an excel file.
"""
import regex
import logging
import itertools
import schedula as sh
from collections.abc import Iterable

log = logging.getLogger(__name__)


def _clone_excel(file_name):
    from urllib.error import URLError
    import openpyxl
    try:
        from urllib.request import urlopen
        book = openpyxl.load_workbook(urlopen(file_name))
    except (ValueError, URLError):
        with open(file_name, 'rb') as file:
            book = openpyxl.load_workbook(file)
    import io
    import pandas as pd
    fd = io.BytesIO()
    # noinspection PyTypeChecker
    writer = pd.ExcelWriter(
        fd, engine='openpyxl', optimized_write=True, write_only=True
    )
    writer.book = book
    writer.sheets.update(dict((ws.title, ws) for ws in book.worksheets))
    return writer, fd


def _sort_sheets(x):
    x = x[0]
    imp = ['summary', 'dice', 'graphs', 'plan', 'nedc_h', 'nedc_l', 'wltp_h',
           'wltp_l', 'wltp_m', 'wltp_p', 'prediction', 'calibration', 'input',
           'mt', 'pa', 'ts']

    w = ()
    for i, k in enumerate(imp):
        if k in x:
            w = (i,) + _sort_sheets((x.replace(k, ''),))[0]
            break
    return w or (100,), x


def _multi_index_df2excel(writer, shname, df, index=True, **kw):
    try:
        df.to_excel(writer, shname, index=index, **kw)
    except NotImplementedError as ex:
        import pandas as pd
        if not index and isinstance(df.columns, pd.MultiIndex):
            kw = kw.copy()
            if kw.pop('header', True):
                header = pd.DataFrame([c for c in df.columns]).T
                header.to_excel(writer, shname, index=False, header=False, **kw)
                kw['startrow'] = kw.get('startrow', 0) + header.shape[0]
            values = pd.DataFrame(df.values)
            values.to_excel(writer, shname, index=False, header=False, **kw)
        else:
            raise ex


def _index_levels(index):
    # noinspection PyBroadException
    try:
        return len(index.levels)
    except Exception:
        return 1


# noinspection PyUnusedLocal
def _get_corner(df, startcol=0, startrow=0, index=True, header=True):
    if header:
        i = _index_levels(df.columns)
        startrow += i

        import pandas as pd
        if index and isinstance(df.columns, pd.MultiIndex):
            startrow += 1

    if index:
        i = _index_levels(df.index)
        startcol += i
    return startrow, startcol


def _convert_index(k):
    if not isinstance(k, Iterable):
        k = (str(k),)
    elif isinstance(k, str):
        k = (k,)
    return k


def _rangename2d(rlo, clo, rhi, chi):
    import xlrd
    return "%s:%s" % (xlrd.cellnameabs(rlo, clo), xlrd.cellnameabs(rhi, chi))


def _ranges_by_col_row(df, startrow, startcol):
    for row, i in enumerate(df.index, start=startrow):
        i = _convert_index(i)
        for col, c in enumerate(df.columns, start=startcol):
            yield i + _convert_index(c), _rangename2d(row, col, row, col)


def _ranges_by_col(df, startrow, startcol):
    for col, (k, v) in enumerate(df.items(), start=startcol):
        yield k, _rangename2d(startrow, col, startrow + len(v) - 1, col)


def _ranges_by_row(df, startrow, startcol):
    for row, (k, v) in enumerate(df.iterrows(), start=startrow):
        yield k, _rangename2d(row, startcol, row, startcol + len(v) - 1)


def _add_named_ranges(df, writer, shname, startrow, startcol, named_ranges, k0):
    ref = '!'.join([shname, '%s'])
    # noinspection PyBroadException
    try:
        define_name = writer.book.define_name

        def _create_named_range(ref_n, ref_r):
            define_name(ref % ref_n, ref % ref_r)
    except Exception:  # Use other pkg.
        define_name = writer.book.create_named_range
        scope = writer.book.index(writer.sheets[shname])

        def _create_named_range(ref_n, ref_r):
            define_name(ref_n, value=ref % ref_r, scope=scope)

    tag = ()
    if hasattr(df, 'name'):
        tag += (df.name,)

    it = ()

    if 'rows' in named_ranges and 'columns' in named_ranges:
        it += (_ranges_by_col_row(df, startrow, startcol),)
    elif 'columns' in named_ranges:
        it += (_ranges_by_col(df, startrow, startcol),)
    elif 'rows' in named_ranges:
        it += (_ranges_by_row(df, startrow, startcol),)

    for k, range_ref in itertools.chain(*it):
        k = _convert_index(k)
        if k:
            try:
                k = tag + k[k0:]
                _create_named_range(_ref_name(*k), range_ref)
            except TypeError:
                pass


def _df2excel(writer, shname, df, k0=0, named_ranges=('columns', 'rows'), **kw):
    import pandas as pd
    if isinstance(df, pd.DataFrame) and not df.empty:
        _multi_index_df2excel(writer, shname, df, **kw)

        startrow, startcol = _get_corner(df, **kw)

        if named_ranges:
            _add_named_ranges(df, writer, shname, startrow, startcol,
                              named_ranges, k0)

        return startrow, startcol


def _write_sheets(writer, sheet_name, data, down=True, **kw):
    import pandas as pd
    if isinstance(data, pd.DataFrame):
        return _df2excel(writer, sheet_name, data, **kw)
    else:
        corner = None
        for d in data:
            corner = _write_sheets(writer, sheet_name, d, down=not down, **kw)
            if corner:
                if down:
                    kw['startrow'] = d.shape[0] + corner[0] + 2
                else:
                    kw['startcol'] = d.shape[1] + corner[1] + 2
        return corner


def _sheet_name(tags):
    return '.'.join(tags)


def _ref_name(*names):
    return '_{}'.format(regex.sub(r"[\W]", "_", '.'.join(names)))


def _data_ref(ref):
    return '%s!%s' % (_sheet_name(ref[:-1]), _ref_name(ref[-1]))


def _chart2excel(writer, sheet, charts):
    import xlrd
    from openpyxl.chart import ScatterChart, Series

    sn = writer.book.sheetnames
    named_ranges = {'%s!%s' % (sn[d.localSheetId], d.name): d.value
                    for d in writer.book.defined_names.definedName}
    m, h, w = 3, 7.94, 13.55

    for i, (k, v) in enumerate(sorted(charts.items())):
        chart = ScatterChart()
        chart.height = h
        chart.width = w
        _map = {
            ('title', 'name'): ('title',),
            ('y_axis', 'name'): ('y_axis', 'title'),
            ('x_axis', 'name'): ('x_axis', 'title'),
        }
        _filter = {
            ('legend', 'position'): lambda x: x[0],
        }
        it = {s: _filter[s](o) if s in _filter else o
              for s, o in sh.stack_nested_keys(v['set'])}

        for s, o in sh.map_dict(_map, it).items():
            c = chart
            for j in s[:-1]:
                c = getattr(c, j)
            setattr(c, s[-1], o)

        for s in v['series']:
            xvalues = named_ranges[_data_ref(s['x'])]
            values = named_ranges[_data_ref(s['y'])]
            series = Series(values, xvalues, title=s['label'])
            chart.series.append(series)

        n = int(i / m)
        j = i - n * m

        sheet.add_chart(chart, xlrd.cellname(15 * j, 8 * n))


def write_to_excel(dfs, output_template):
    """
    Writes DataFrames to excel.

    :param dfs:
        DataFrames of vehicle output report.
    :type dfs: dict[str, pandas.DataFrame]

    :param output_template:
        Template output.
    :type output_template: str

    :return:
        Excel output file.
    :rtype: io.BytesIO
    """
    log.debug(
        'Writing into xl-file based on template(%s)...', output_template
    )
    writer, fd = _clone_excel(output_template)

    calculate_sheets, charts = sorted(writer.sheets), []
    for k, v in sorted(dfs.items(), key=_sort_sheets):
        if not k.startswith('graphs.'):
            down = True
            if k.endswith('pa'):
                kw = {'named_ranges': ('rows',), 'index': True, 'k0': 1}
            elif k.endswith('ts'):
                kw = {'named_ranges': ('columns',), 'index': False, 'k0': 1}
            elif k.endswith('proc_info'):
                down = False
                kw = {'named_ranges': ()}
            else:
                kw = {}
            _write_sheets(writer, k, v, down=down, **kw)
        else:
            try:
                sheet = writer.book.add_worksheet(k)
            except AttributeError:
                sheet = writer.book.create_sheet(title=k)
            charts.append((sheet, v))

    for sheet, v in charts:
        _chart2excel(writer, sheet, v)

    writer.save()
    return fd



# -*- coding: utf-8 -*-
#
# Copyright 2015-2022 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
"""
Functions and `dsp` model to model the engine coolant temperature.
"""
import numpy as np
import schedula as sh
from co2mpas.defaults import dfl

dsp = sh.BlueDispatcher(
    name='thermal', description='Models the engine thermal behaviour.'
)


def _derivative(times, temp):
    import scipy.misc as sci_misc
    import scipy.interpolate as sci_itp
    par = dfl.functions.calculate_engine_temperature_derivatives
    func = sci_itp.InterpolatedUnivariateSpline(times, temp, k=1)
    return sci_misc.derivative(func, times, dx=par.dx, order=par.order)


@sh.add_function(dsp, outputs=['engine_temperatures', 'temperature_shift'])
def calculate_engine_temperatures(
        engine_thermostat_temperature, times, engine_coolant_temperatures,
        on_engine):
    from statsmodels.nonparametric.smoothers_lowess import lowess
    from syncing.model import dsp
    par = dfl.functions.calculate_engine_temperature_derivatives
    temp = lowess(
        engine_coolant_temperatures, times, is_sorted=True,
        frac=par.tw * len(times) / (times[-1] - times[0]) ** 2, missing='none'
    )[:, 1].ravel()
    i = np.searchsorted(temp, engine_thermostat_temperature)
    shifts = dsp({
        'reference_name': 'ref', 'data': {
            'ref': {'x': times[:i], 'y': on_engine[:i]},
            'data': {'x': times[:i], 'y': _derivative(times[:i], temp[:i])}
        }
    }, ['shifts'])['shifts']
    if not (-20 <= shifts['data'] <= 30):
        shifts['data'] = 0
    sol = dsp({
        'shifts': shifts, 'reference_name': 'ref', 'data': {
            'ref': {'x': times},
            'data': {'x': times, 'y': temp}
        },
        'interpolation_method': 'cubic'
    })
    return sol['resampled']['data']['y'], shifts['data']


@sh.add_function(dsp, outputs=['engine_temperature_derivatives'])
def calculate_engine_temperature_derivatives(
        times, engine_temperatures):
    """
    Calculates the derivative of the engine temperature [°C/s].

    :param times:
        Time vector [s].
    :type times: numpy.array

    :param engine_coolant_temperatures:
        Engine coolant temperature vector [°C].
    :type engine_coolant_temperatures: numpy.array

    :return:
        Derivative of the engine temperature [°C/s].
    :rtype: numpy.array
    """
    return _derivative(times, engine_temperatures)


@sh.add_function(dsp, outputs=['max_engine_coolant_temperature'])
def identify_max_engine_coolant_temperature(engine_coolant_temperatures):
    """
    Identifies maximum engine coolant temperature [°C].

    :param engine_coolant_temperatures:
        Engine coolant temperature vector [°C].
    :type engine_coolant_temperatures: numpy.array

    :return:
        Maximum engine coolant temperature [°C].
    :rtype: float
    """
    return engine_coolant_temperatures.max()


@sh.add_function(dsp, outputs=['engine_temperature_regression_model'])
def calibrate_engine_temperature_regression_model(
        engine_thermostat_temperature, engine_temperatures, velocities,
        engine_temperature_derivatives, on_engine, engine_speeds_out,
        accelerations, after_treatment_warm_up_phases):
    """
    Calibrates an engine temperature regression model to predict engine
    temperatures.

    :param engine_thermostat_temperature:
        Engine thermostat temperature [°C].
    :type engine_thermostat_temperature: float

    :param on_engine:
        If the engine is on [-].
    :type on_engine: numpy.array

    :param engine_temperature_derivatives:
        Derivative of the engine temperature [°C/s].
    :type engine_temperature_derivatives: numpy.array

    :param engine_coolant_temperatures:
        Engine coolant temperature vector [°C].
    :type engine_coolant_temperatures: numpy.array

    :param velocities:
        Velocity [km/h].
    :type velocities: numpy.array

    :param engine_speeds_out:
        Engine speed [RPM].
    :type engine_speeds_out: numpy.array

    :param accelerations:
        Acceleration vector [m/s2].
    :type accelerations: numpy.array

    :return:
        The calibrated engine temperature regression model.
    :rtype: callable
    """
    from ._thermal import ThermalModel
    return ThermalModel(engine_thermostat_temperature).fit(
        engine_temperatures, engine_temperature_derivatives, on_engine,
        velocities, engine_speeds_out, accelerations,
        after_treatment_warm_up_phases
    )


@sh.add_function(dsp, outputs=['engine_temperatures'])
def predict_engine_temperatures(
        engine_temperature_regression_model, times, on_engine, velocities,
        engine_speeds_out, accelerations, after_treatment_warm_up_phases,
        initial_engine_temperature, max_engine_coolant_temperature):
    """
    Predicts the engine temperature [°C].

    :param engine_temperature_regression_model:
        Engine temperature regression engine_temperature_regression_model.
    :type engine_temperature_regression_model: callable

    :param times:
        Time vector [s].
    :type times: numpy.array

    :param accelerations:
        Acceleration vector [m/s2].
    :type accelerations: numpy.array

    :param on_engine:
        If the engine is on [-].
    :type on_engine: numpy.array

    :param velocities:
        Velocity [km/h].
    :type velocities: numpy.array

    :param engine_speeds_out:
        Engine speed [RPM].
    :type engine_speeds_out: numpy.array

    :param initial_engine_temperature:
        Engine initial temperature [°C]
    :type initial_engine_temperature: float

    :param max_engine_coolant_temperature:
        Maximum engine coolant temperature [°C].
    :type max_engine_coolant_temperature: float

    :return:
        Engine coolant temperature vector [°C].
    :rtype: numpy.array
    """
    return engine_temperature_regression_model(
        times, on_engine, velocities, engine_speeds_out, accelerations,
        after_treatment_warm_up_phases, max_temp=max_engine_coolant_temperature,
        initial_temperature=initial_engine_temperature
    )


@sh.add_function(dsp, outputs=['engine_coolant_temperatures'])
def calculate_engine_coolant_temperatures(
        times, engine_temperatures, temperature_shift):
    from syncing.model import dsp
    sol = dsp({
        'shifts': {'data': -temperature_shift}, 'data': {
            'ref': {'x': times},
            'data': {'x': times, 'y': engine_temperatures}
        },
        'reference_name': 'ref', 'interpolation_method': 'cubic'
    })
    return sol['resampled']['data']['y']


# noinspection PyPep8Naming
@sh.add_function(
    dsp, inputs_kwargs=True, outputs=['engine_thermostat_temperature']
)
@sh.add_function(
    dsp, function_id='identify_engine_thermostat_temperature_v1',
    weight=sh.inf(11, 100), outputs=['engine_thermostat_temperature']
)
def identify_engine_thermostat_temperature(
        idle_engine_speed, times, accelerations, engine_coolant_temperatures,
        engine_speeds_out, gear_box_powers_out=None):
    """
    Identifies thermostat engine temperature and its limits [°C].

    :param idle_engine_speed:
        Engine speed idle median and std [RPM].
    :type idle_engine_speed: (float, float)

    :param times:
        Time vector [s].
    :type times: numpy.array

    :param engine_coolant_temperatures:
        Engine coolant temperature vector [°C].
    :type engine_coolant_temperatures: numpy.array

    :param gear_box_powers_out:
        Gear box power out vector [kW].
    :type gear_box_powers_out: numpy.array

    :param engine_speeds_out:
        Engine speed [RPM].
    :type engine_speeds_out: numpy.array

    :param accelerations:
        Acceleration vector [m/s2].
    :type accelerations: numpy.array

    :return:
        Engine thermostat temperature [°C].
    :rtype: float
    """
    args = engine_speeds_out, accelerations
    if gear_box_powers_out is not None:
        args += gear_box_powers_out,
    from ._thermal import _build_samples, _XGBRegressor
    X, Y = _build_samples(
        _derivative(times, engine_coolant_temperatures),
        engine_coolant_temperatures, *args
    )
    X, Y = np.column_stack((Y, X[:, 1:])), X[:, 0]
    t_max, t_min = Y.max(), Y.min()
    b = (t_max - (t_max - t_min) / 3) <= Y

    # noinspection PyArgumentEqualDefault
    model = _XGBRegressor(
        random_state=0, objective='reg:squarederror'
    ).fit(X[b], Y[b])
    ratio = np.arange(1, 1.5, 0.1) * idle_engine_speed[0]
    spl = np.zeros((len(ratio), X.shape[1]))
    spl[:, 1] = ratio
    # noinspection PyTypeChecker
    return float(np.median(model.predict(spl)))


@sh.add_function(dsp, outputs=['engine_thermostat_temperature_window'])
def identify_engine_thermostat_temperature_window(
        engine_thermostat_temperature, engine_coolant_temperatures):
    """
    Identifies thermostat engine temperature limits [°C].

    :param engine_thermostat_temperature:
        Engine thermostat temperature [°C].
    :type engine_thermostat_temperature: float

    :param engine_coolant_temperatures:
        Engine coolant temperature vector [°C].
    :type engine_coolant_temperatures: numpy.array

    :return:
        Thermostat engine temperature limits [°C].
    :rtype: float, float
    """

    thr = engine_thermostat_temperature
    # noinspection PyTypeChecker
    std = np.sqrt(np.mean((engine_coolant_temperatures - thr) ** 2))
    return thr - std, thr + std


@sh.add_function(dsp, outputs=['initial_engine_temperature'])
def identify_initial_engine_temperature(engine_temperatures):
    """
    Identifies initial engine temperature [°C].

    :param engine_coolant_temperatures:
        Engine coolant temperature vector [°C].
    :type engine_coolant_temperatures: numpy.array

    :return:
        Initial engine temperature [°C].
    :rtype: float
    """
    return float(engine_temperatures[0])
